from __future__ import annotations

import json
import re
import sys
import textwrap
from abc import ABC, abstractmethod
from datetime import datetime, timezone, date
from functools import lru_cache
from json import JSONDecodeError
from typing import (
    Union,
    Any,
    Optional,
    Callable,
    Type,
    Sequence,
    Dict,
    List,
    Set,
    cast,
    Tuple,
    Iterable,
    TypeVar,
    NamedTuple,
)

import yaml
from attrs import define, frozen
from dateutil.parser import parse
from jsons import set_deserializer, set_serializer
from networkx import MultiDiGraph
from parsy import regex, string, Parser

from resotocore.compat import remove_suffix
from resotocore.model.transform_kind_convert import converters
from resotocore.model.typed_model import from_js, to_js
from resotocore.types import Json, JsonElement, ValidationResult, ValidationFn, EdgeType
from resotocore.util import if_set, utc, duration, first
from resotolib.core.model_check import check_overlap_for
from resotolib.durations import duration_parser, DurationRe
from resotolib.parse_util import make_parser, variable_dp_backtick, dot_dp
from resotolib.utils import is_env_var_string

T = TypeVar("T")


def check_type_fn(t: type, type_name: str) -> ValidationFn:
    def check_type(x: Any) -> ValidationResult:
        if isinstance(x, t):
            return None
        else:
            raise AttributeError(f"Expected type {type_name} but got {type(x).__name__}")

    return check_type


def check_fn(x: Optional[Any], func: Callable[[Any, Any], Optional[Any]], message: str) -> Optional[ValidationFn]:
    def check_single(value: Any) -> ValidationResult:
        if func(x, value):
            return None
        else:
            raise AttributeError(f">{value}< {message}")

    return None if x is None else check_single


def validate_fn(*fns: Optional[ValidationFn]) -> ValidationFn:
    defined = list(filter(lambda x: x is not None, fns))

    def always_valid(_: Any) -> ValidationResult:
        return None

    def check_defined(value: Any) -> ValidationResult:
        for fn in defined:
            res = fn(value)  # type: ignore
            if res is not None:
                return res
        return None

    return check_defined if defined else always_valid


@define(order=True, hash=True, frozen=True)
class SyntheticProperty:
    path: List[str]
    """
    A synthetic property does not exist in the underlying data model.
    It is defined by a function on an existing other property.
    Example: age is a duration defined on ctime which is a datetime.
             the function is age=now-ctime.
    """


@define(order=True, hash=True, eq=True)
class Property:
    name: str
    kind: str
    required: bool = False
    synthetic: Optional[SyntheticProperty] = None
    description: Optional[str] = None
    metadata: Optional[Json] = None

    def __attrs_post_init__(self) -> None:
        assert self.synthetic is None or not self.required, "Synthetic properties can not be required!"

    def resolve(self, model: Dict[str, Kind]) -> Kind:
        return Property.parse_kind(self.kind, model)

    def meta(self, name: str, clazz: Type[T]) -> Optional[T]:
        meta = self.metadata
        return v if meta is not None and (v := meta.get(name)) is not None and isinstance(v, clazz) else None

    def meta_get(self, name: str, clazz: Type[T], default: T) -> T:
        return self.meta(name, clazz) or default

    @staticmethod
    def parse_kind(name: str, model: Dict[str, Kind]) -> Kind:
        def kind_by_name(kind_name: str) -> Kind:
            if kind_name not in model:
                raise AttributeError(f"Property kind is not known: {kind_name}. Have you registered it?")
            return model[kind_name]

        simple_kind_parser = regex("[A-Za-z][A-Za-z0-9_.]*").map(kind_by_name)
        bracket_parser = string("[]")
        dict_string_parser = string("dictionary[")
        comma_parser = regex("\\s*,\\s*")
        bracket_r = string("]")

        @make_parser
        def array_parser() -> Parser:
            inner = yield dictionary_parser | simple_kind_parser
            brackets = yield bracket_parser.times(1, cast(int, float("inf")))
            return ArrayKind.mk_array(inner, len(brackets))

        @make_parser
        def dictionary_parser() -> Parser:
            yield dict_string_parser
            key_kind = cast(Kind, (yield simple_kind_parser))
            yield comma_parser
            value_kind = yield array_parser | dictionary_parser | simple_kind_parser
            yield bracket_r
            return DictionaryKind(key_kind, value_kind)

        return (array_parser | dictionary_parser | simple_kind_parser).parse(name)  # type: ignore

    @staticmethod
    def any_prop() -> Property:
        return Property("any", "any")


# Split a variable path into its path parts.
# foo.bla -> [foo, bla]
# foo.`bla.bar` -> [foo, bla.bar]
prop_path_parser = (regex("[^`.]+") | variable_dp_backtick).sep_by(dot_dp)
array_index_re = re.compile(r"\[(\d+|\*)]")


class PropertyPath:
    @staticmethod
    def from_path(path: str) -> PropertyPath:
        # remove index accesses from the path (e.g. [23] -> "[]", [*] -> "[]")
        no_index = array_index_re.sub("[]", path)
        return PropertyPath(prop_path_parser.parse(no_index), no_index)

    def __init__(self, path: Sequence[Optional[str]], str_rep: Optional[str] = None):
        self.path = path
        self.path_str = str_rep if str_rep else ".".join(a if a else "" for a in self.path)

    @property
    def root(self) -> bool:
        return not bool(self.path)

    def child(self, part: Optional[str]) -> PropertyPath:
        update = list(self.path)
        update.append(part)
        return PropertyPath(update)

    @property
    def last_part(self) -> Optional[str]:
        return self.path[-1] if self.path else None

    def same_as(self, other: PropertyPath) -> bool:
        """
        Checks if the given path is the same this path.
        Note: the path may include "holes" marked as None.
              The holes mean positions that have to be ignored.
              [A,B,C] same_as [A,None,C] same_as [A,None,None] same_as [None,None,None]
        """
        if len(other.path) == len(self.path):
            for left, right in zip(self.path, other.path):
                if left is not None and right is not None and left != right:
                    return False
            return True
        else:
            return False

    def __repr__(self) -> str:
        return self.path_str

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, PropertyPath) and other.path_str == self.path_str

    def __hash__(self) -> int:
        return hash(self.path_str)


EmptyPath = PropertyPath([], "")


@define(order=True, hash=True, frozen=True)
class ResolvedProperty:
    # The path of the resolved property in a complex kind
    path: PropertyPath
    # The metadata of the property (name etc.)
    prop: Property
    # the resolved kind of this property
    kind: Kind

    @property
    def simple_kind(self) -> SimpleKind:
        return self.kind if isinstance(self.kind, SimpleKind) else AnyKind()


class Kind(ABC):
    def __init__(self, fqn: str):
        self.fqn = fqn

    def __eq__(self, other: object) -> bool:
        return self.__dict__ == other.__dict__ if isinstance(other, Kind) else False

    def __hash__(self) -> int:
        return hash(self.fqn)

    def coerce(self, value: JsonElement, **kwargs: bool) -> JsonElement:
        coerced = self.coerce_if_required(value, **kwargs)
        return coerced or value

    @abstractmethod
    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[JsonElement]:
        """
        Coerces the given value to this kind.
        - if obj is already correct: return None
        - if obj can not be coerced: return None
        - if obj can be coerced: return the value
        """

    @abstractmethod
    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        """
        Validate given object against definition:
        - if obj is valid: return None
        - if obj is not valid and can be coerced: return the coerced value.
        - if obj is not valid and can not be coerced: raise an exception
        """

    def resolve(self, model: Dict[str, Kind]) -> None:
        pass

    def kind_hierarchy(self) -> Set[str]:
        return {self.fqn}

    def package(self) -> Optional[str]:
        return self.fqn.rsplit(".", 1)[0] if "." in self.fqn else None

    # noinspection PyUnusedLocal
    @staticmethod
    def from_json(js: Json, _: type = object, **kwargs: object) -> Kind:
        if "inner" in js:
            inner = Kind.from_json(js["inner"])
            return ArrayKind(inner)
        elif "fqn" in js and "runtime_kind" in js and js["runtime_kind"] in simple_kind_to_type:
            fqn = js["fqn"]
            rk = js["runtime_kind"]
            if "source_fqn" in js and "converter" in js and "reverse_order" in js:
                return TransformKind(fqn, rk, js["source_fqn"], js["converter"], js["reverse_order"])
            elif rk == "string":
                minimum = js.get("min_length")
                maximum = js.get("max_length")
                p = js.get("pattern")
                e = js.get("enum")
                return StringKind(fqn, minimum, maximum, p, e)
            elif rk in ("int32", "int64", "float", "double"):
                minimum = js.get("minimum")
                maximum = js.get("maximum")
                e = js.get("enum")
                return NumberKind(fqn, rk, minimum, maximum, e)
            elif rk == "datetime":
                return DateTimeKind(fqn)
            elif rk == "date":
                return DateKind(fqn)
            elif rk == "duration":
                return DurationKind(fqn)
            elif rk == "boolean":
                return BooleanKind(fqn)
            else:
                raise TypeError(f"Unhandled runtime kind: {rk}")
        elif js.get("fqn") == "any" and js.get("runtime_kind") == "any":
            return AnyKind()
        elif "fqn" in js and ("properties" in js or "bases" in js):
            props = list(map(lambda p: from_js(p, Property), js.get("properties", [])))
            bases: List[str] = js.get("bases", [])
            allow_unknown_props = js.get("allow_unknown_props", False)
            successor_kinds = js.get("successor_kinds")
            aggregate_root = js.get("aggregate_root", True)
            metadata = js.get("metadata")
            return ComplexKind(js["fqn"], bases, props, allow_unknown_props, successor_kinds, aggregate_root, metadata)
        else:
            raise JSONDecodeError("Given type can not be read.", json.dumps(js), 0)

    def sort_json(self, js: Json) -> Json:
        """
        Sort the given json element according to the kind definition.
        A model which define properties in a specific order, the json object also uses the same order.
        This will walk the complete json structure.
        """
        if isinstance(js, dict) and isinstance(self, ComplexKind):
            prop_order = {rp.name: idx for idx, rp in enumerate(self.all_props())}
            sub = {
                k: self.property_kind_of(k, AnyKind()).sort_json(v) if isinstance(v, dict) else v for k, v in js.items()
            }
            # a property which is unknown is moved to the end with no specific order
            return dict(sorted(sub.items(), key=lambda k: prop_order.get(k[0], sys.maxsize)))
        elif isinstance(js, dict):
            # noinspection PyTypeChecker
            return dict(sorted(js.items()))
        else:
            return js

    def nested_complex_kinds(self) -> List[ComplexKind]:
        if isinstance(self, ComplexKind):
            return [self]
        elif isinstance(self, ArrayKind):
            return self.inner.nested_complex_kinds()
        elif isinstance(self, DictionaryKind):
            return self.key_kind.nested_complex_kinds() + self.value_kind.nested_complex_kinds()
        else:
            return []


def should_skip_env_var_str(obj: Any, kwargs: Dict[str, bool]) -> bool:
    return kwargs.get("config_context", False) and is_env_var_string(obj)


simple_kind_to_type: Dict[str, Type[Union[str, int, float, bool]]] = {
    "string": str,
    "int32": int,
    "int64": int,
    "float": float,
    "double": float,
    "boolean": bool,
    "date": str,
    "datetime": str,
    "duration": str,
}


class SimpleKind(Kind, ABC):
    def __init__(self, fqn: str, runtime_kind: str, reverse_order: bool = False):
        super().__init__(fqn)
        self.runtime_kind = runtime_kind
        self.reverse_order = reverse_order

    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[JsonElement]:
        """
        Take a user defined value and transform it into a machine queryable value.
        Example:
            - "10s" as string -> "10s"
            - "10s" as boolean -> false
            - "10s" as duration -> "10s"
            - "10s" as date -> now + 10 seconds (depending on local time)
        :param value: the value from the user
        :return: the coerced value from the system
        """
        return None

    def as_json(self) -> Json:
        return {"fqn": self.fqn, "runtime_kind": self.runtime_kind}

    # noinspection PyUnusedLocal
    @staticmethod
    def to_json(obj: SimpleKind, **kw_args: object) -> Json:
        return obj.as_json()


class AnyKind(SimpleKind):
    def __init__(self) -> None:
        super().__init__("any", "any")

    def __new__(cls) -> AnyKind:
        if cls.__singleton is None:
            cls.__singleton = super(AnyKind, cls).__new__(cls)
        return cls.__singleton

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        return None

    __singleton: Optional[AnyKind] = None


class StringKind(SimpleKind):
    def __init__(
        self,
        fqn: str,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        pattern: Optional[str] = None,
        enum: Optional[Set[str]] = None,
    ):
        super().__init__(fqn, "string")
        self.min_length = min_length
        self.max_length = max_length
        self.pattern = pattern
        self.pattern_compiled = if_set(pattern, re.compile)
        self.enum = enum
        self.valid_fn = validate_fn(
            check_type_fn(str, "string"),
            check_fn(
                self.pattern_compiled,
                lambda p, obj: p.fullmatch(obj) is not None,
                f"does not conform to regex: {self.pattern}",
            ),
            check_fn(self.enum, lambda x, obj: obj in x, f"should be one of: {self.enum}"),
            check_fn(self.min_length, lambda x, obj: len(obj) >= x, f"does not have minimal length: {self.min_length}"),
            check_fn(self.max_length, lambda x, obj: len(obj) <= x, f"is too long! Allowed: {self.max_length}"),
        )

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        if obj is None:
            return None
        elif should_skip_env_var_str(obj, kwargs):
            return None
        elif isinstance(obj, str):
            return self.valid_fn(obj)
        coerced = self.coerce_if_required(obj, **kwargs)
        if coerced is not None:
            self.valid_fn(coerced)
            return coerced
        else:
            raise AttributeError(f"Expected type {self.runtime_kind} but got {type(obj).__name__}")

    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[str]:
        if value is None:
            return value
        if isinstance(value, str):
            return None
        else:
            return json.dumps(value)

    def as_json(self) -> Json:
        js = super().as_json()
        if self.pattern:
            js["pattern"] = self.pattern
        if self.enum:
            js["enum"] = self.enum
        if self.min_length:
            js["min_length"] = self.min_length
        if self.max_length:
            js["max_length"] = self.max_length
        return js


class NumberKind(SimpleKind):
    def __init__(
        self,
        fqn: str,
        runtime_kind: str,
        minimum: Union[None, float, int] = None,
        maximum: Union[None, float, int] = None,
        enum: Optional[Set[Union[float, int]]] = None,
    ):
        super().__init__(fqn, runtime_kind)
        self.minimum = minimum
        self.maximum = maximum
        self.enum = enum
        self.valid_fn = validate_fn(
            check_type_fn(int, "int") if runtime_kind == "int" else self.check_float,
            check_fn(self.enum, lambda x, obj: obj in x, f"should be one of: {self.enum}"),
            check_fn(self.minimum, lambda x, obj: obj >= x, f"should be greater or equals than: {self.minimum}"),
            check_fn(self.maximum, lambda x, obj: obj <= x, f"should be smaller or equals than: {self.maximum}"),
        )

    @staticmethod
    def check_float(obj: Any) -> ValidationResult:
        if isinstance(obj, (int, float)):
            return None
        else:
            raise AttributeError(f"Expected number but got {obj}")

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        if obj is None:
            return None
        elif should_skip_env_var_str(obj, kwargs):
            return None
        elif isinstance(obj, (int, float)):
            return self.valid_fn(obj)
        coerced = self.coerce_if_required(obj, **kwargs)
        if coerced is not None:
            self.valid_fn(coerced)
            return coerced
        else:
            raise AttributeError(f"Expected type {self.runtime_kind} but got {type(obj).__name__}")

    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[Union[int, float]]:
        if value is None:
            return value
        elif isinstance(value, (int, float)):
            return None
        else:
            try:
                return int(value) if self.runtime_kind in ("int32", "int64") else float(value)  # type: ignore
            except ValueError:
                return None

    def as_json(self) -> Json:
        js = super().as_json()
        if self.enum:
            js["enum"] = self.enum
        if self.minimum:
            js["minimum"] = self.minimum
        if self.maximum:
            js["maximum"] = self.maximum
        return js


class BooleanKind(SimpleKind):
    def __init__(self, fqn: str):
        super().__init__(fqn, "boolean")
        self.valid_fn = check_type_fn(bool, "boolean")

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        if obj is True or obj is False or obj is None:
            return None
        elif should_skip_env_var_str(obj, kwargs):
            return None
        coerced = self.coerce_if_required(obj, **kwargs)
        if coerced is not None:
            return coerced
        else:
            raise AttributeError(f"Expected type boolean but got {type(obj).__name__}")

    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[bool]:
        if value is None:
            return value
        elif isinstance(value, bool):
            return None
        elif isinstance(value, str) and value.lower() in ("true", "yes", "on"):
            return True
        elif isinstance(value, str) and value.lower() in ("false", "no", "off"):
            return False
        else:
            return None


class DurationKind(SimpleKind):
    def __init__(self, fqn: str):
        super().__init__(fqn, "duration")
        self.valid_fn = validate_fn(check_type_fn(str, "duration"), self.check_duration)

    def check_duration(self, v: Any) -> None:
        if not DurationRe.fullmatch(v):
            raise AttributeError(f"Wrong format for duration: {v}. Examples: 1yr, 3mo, 3d4h3min1s, 3days and 2hours")

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        if should_skip_env_var_str(obj, kwargs):
            return None
        return self.valid_fn(obj)

    def coerce_if_required(self, value: Any, **kwargs: bool) -> Optional[str]:
        try:
            return f"{int(duration_parser.parse(value))}s" if kwargs.get("normalize", True) else None
        except Exception:
            return None


class DateTimeKind(SimpleKind):
    Format = "%Y-%m-%dT%H:%M:%SZ"
    DateTimeRe = re.compile("\\d{4}-\\d{2}-\\d{2}T\\d{2}:\\d{2}:\\d{2}Z")

    def __init__(self, fqn: str):
        super().__init__(fqn, "datetime")
        self.valid_fn = validate_fn(check_type_fn(str, "datetime"), self.check_datetime)

    @staticmethod
    def parse_datetime(date_string: str) -> Optional[datetime]:
        try:
            return datetime.strptime(date_string, DateTimeKind.Format)
        except ValueError:
            return None

    @staticmethod
    def check_datetime(obj: Any) -> ValidationResult:
        def parse_datetime() -> str:
            parsed = datetime.fromisoformat(str(obj))
            utc_parsed = datetime.fromtimestamp(parsed.timestamp(), tz=timezone.utc)
            return utc_parsed.strftime(DateTimeKind.Format)

        return None if DateTimeKind.DateTimeRe.fullmatch(obj) else parse_datetime()

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        if should_skip_env_var_str(obj, kwargs):
            return None
        return self.valid_fn(obj)

    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[str]:
        try:
            if value is None:
                return value
            elif not isinstance(value, str):
                return None
            elif self.DateTimeRe.fullmatch(value):
                return None
            elif DurationRe.fullmatch(value):
                return self.from_duration(value)
            else:
                return self.from_datetime(value)
        except Exception:
            return None

    @staticmethod
    def from_datetime(value: str) -> str:
        try:
            dt = datetime.fromisoformat(value)
        except Exception:
            dt = parse(value)
        if (
            not dt.tzinfo
            or dt.tzinfo.utcoffset(None) is None
            or dt.tzinfo.utcoffset(None).total_seconds() != 0  # type: ignore
        ):
            dt = dt.astimezone(timezone.utc)
        return dt.strftime(DateTimeKind.Format)

    @staticmethod
    def from_duration(value: str, now: datetime = utc()) -> str:
        # in case of duration, compute the timestamp as: now + duration
        delta = duration(value)
        instant = now + delta
        return instant.strftime(DateTimeKind.Format)


class DateKind(SimpleKind):
    Format = "%Y-%m-%d"
    DateRe = re.compile("\\d{4}-\\d{2}-\\d{2}")

    def __init__(self, fqn: str):
        super().__init__(fqn, "date")
        self.valid_fn = validate_fn(check_type_fn(str, "date"), self.check_date)

    @staticmethod
    def check_date(obj: Any) -> ValidationResult:
        return None if DateKind.DateRe.fullmatch(obj) else date.fromisoformat(obj)

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        if should_skip_env_var_str(obj, kwargs):
            return None
        return self.valid_fn(obj)

    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[str]:
        try:
            if value is None:
                return value
            elif not isinstance(value, str):
                return None
            elif DurationRe.fullmatch(value):
                # in case of duration, compute the timestamp as: today + duration
                delta = duration(value)
                at = date.today() + delta
                return at.isoformat()
            else:
                return parse(value).date().strftime(DateKind.Format)
        except Exception:
            return None


class TransformKind(SimpleKind):
    """
    Transform kinds can be used to derive attributes in a complex kind from other attributes.
    It is important, that the transformed attribute does not exist in the original complex kind!
    It is a SimpleKind, since it can be queried directly as if it would be available as part of the json.

    A transformed kind takes a source value of kind source_kind and transforms it using a function
    into the destination kind.

    :param fqn: the fully qualified name of this kind.
    :param source_fqn: the underlying runtime kind.
    :param destination_fqn: the destination kind that is used in the data store.
    :param converter: name of converter. See transform_kind_convert.py for a dict of possible converter names.
    """

    def __init__(self, fqn: str, source_fqn: str, destination_fqn: str, converter: str, reverse_order: bool):
        # note: source_fqn and runtime_kind are considered the same.
        # the synthetic property does not introduce a new type, but translates types.
        super().__init__(fqn, source_fqn, reverse_order)
        self.destination_fqn: str = destination_fqn
        self.source_kind: Optional[SimpleKind] = None
        self.destination_kind: Optional[SimpleKind] = None
        self.converter = converter
        self.source_to_destination, self.destination_to_source = converters[converter]

    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[JsonElement]:
        if value is None:
            return None
        elif self.source_kind:
            coerced_source = self.source_kind.coerce(value)
            real = self.source_to_destination(coerced_source)  # type: ignore
            return real
        else:
            return None

    def transform(self, value: Any) -> Any:
        if value is None:
            return None
        elif self.destination_kind:
            real_coerced = self.destination_kind.coerce(value)
            synth = self.destination_to_source(real_coerced)  # type: ignore
            return synth
        else:
            raise AttributeError(f"Synthetic kind is not resolved: {self.fqn}")

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        # this function is called during import for supplied values.
        # synthetic values are never supplied
        raise AttributeError(f"TransformKind {self.fqn} is not allowed to be supplied.")

    def resolve(self, model: Dict[str, Kind]) -> None:
        source = model.get(self.runtime_kind)
        destination = model.get(self.destination_fqn)
        if source and destination and isinstance(source, SimpleKind) and isinstance(destination, SimpleKind):
            source.resolve(model)
            destination.resolve(model)
            self.source_kind = source
            self.destination_kind = destination
        else:
            raise AttributeError(f"Underlying kind not known: {self.destination_fqn}")

    def as_json(self) -> Json:
        return {
            "fqn": self.fqn,
            "runtime_kind": self.runtime_kind,
            "source_fqn": self.destination_fqn,
            "converter": self.converter,
            "reverse_order": self.reverse_order,
        }


class ArrayKind(Kind):
    def __init__(self, inner: Kind):
        super().__init__(f"{inner.fqn}[]")
        self.inner = inner

    def resolve(self, model: Dict[str, Kind]) -> None:
        self.inner.resolve(model)

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        if isinstance(obj, dict):
            # list is expected, but object found
            raise AttributeError("Expected property is a json object not an array!")

        coerced = self.coerce_if_required(obj, **kwargs)
        for elem in coerced or obj:  # type: ignore
            self.inner.check_valid(elem, **kwargs)
        return coerced

    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[List[JsonElement]]:
        has_coerced = False
        if value is None:
            return None
        elif isinstance(value, dict):
            return None
        elif not isinstance(value, list):
            # in case of simple type, we can make it an array
            value = [value]
            has_coerced = True

        def check(item: Any) -> ValidationResult:
            nonlocal has_coerced
            res = self.inner.coerce_if_required(item, **kwargs)
            if res is None:
                return item
            else:
                has_coerced = True
                return res

        mapped = [check(elem) for elem in value]
        return mapped if has_coerced else None

    @staticmethod
    def mk_array(kind: Kind, depth: int) -> Kind:
        return kind if depth == 0 else ArrayKind(ArrayKind.mk_array(kind, depth - 1))


class DictionaryKind(Kind):
    def __init__(self, key_kind: Kind, value_kind: Kind):
        super().__init__(f"dictionary[{key_kind.fqn}, {value_kind.fqn}]")
        self.key_kind = key_kind
        self.value_kind = value_kind

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        if isinstance(obj, dict):
            coerced = self.coerce_if_required(obj, **kwargs)
            for prop, value in (coerced or obj).items():
                part = "key"
                try:
                    self.key_kind.check_valid(prop, **kwargs)
                    part = "value"
                    self.value_kind.check_valid(value, **kwargs)
                except Exception as at:
                    raise AttributeError(f"{part} of {self.fqn} is not valid: {at}") from at
            return coerced
        else:
            raise AttributeError(f"dictionary requires a json object, but got this: {obj}")

    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[Json]:
        if isinstance(value, dict):
            coerced: Json = {}
            has_coerced = False
            for p, v in value.items():
                ck = self.key_kind.coerce_if_required(p, **kwargs)
                cv = self.value_kind.coerce_if_required(v, **kwargs)
                coerced[ck or p] = cv or v  # type: ignore
                if ck is not None or cv is not None:
                    has_coerced = True
            return coerced if has_coerced else None
        else:
            return None

    def resolve(self, model: Dict[str, Kind]) -> None:
        self.key_kind.resolve(model)
        self.value_kind.resolve(model)


class ComplexKind(Kind):
    def __init__(
        self,
        fqn: str,
        bases: List[str],
        properties: List[Property],
        allow_unknown_props: bool = False,
        # EdgeType -> possible list of successor kinds
        successor_kinds: Optional[Dict[EdgeType, List[str]]] = None,
        aggregate_root: bool = True,
        metadata: Optional[Json] = None,
    ):
        super().__init__(fqn)
        self.bases = bases
        self.properties = properties
        self.allow_unknown_props = allow_unknown_props
        self.successor_kinds = successor_kinds or {}
        self.aggregate_root = aggregate_root
        self.metadata = metadata or {}
        self.__prop_by_name = {prop.name: prop for prop in properties}
        self.__resolved = False
        self.__resolved_kinds: Dict[str, Tuple[Property, Kind]] = {}
        self.__resolved_bases: Dict[str, ComplexKind] = {}
        self.__all_props: List[Property] = list(self.properties)
        self.__resolved_hierarchy: Set[str] = {fqn}
        self.__property_by_path: List[ResolvedProperty] = []
        self.__synthetic_props: List[ResolvedProperty] = []

    def resolve(self, model: Dict[str, Kind]) -> None:
        if not self.__resolved:
            self.__resolved = True
            # resolve properties
            for prop in self.properties:
                kind = prop.resolve(model)
                kind.resolve(model)
                self.__resolved_kinds[prop.name] = (prop, kind)

            # make sure all successor kinds can be resolved
            for names in self.successor_kinds.values():
                for name in names or []:
                    if name not in model:
                        raise AttributeError(f"{name} is not a known kind")

            # resolve the hierarchy
            if not self.is_root():
                for base_name in self.bases:
                    base: Kind = model[base_name]
                    base.resolve(model)
                    if isinstance(base, ComplexKind):
                        self.__resolved_bases[base_name] = base
                        self.__resolved_kinds.update(base.__resolved_kinds)
                        self.__all_props = base.__all_props + self.__all_props
                        self.__prop_by_name = {prop.name: prop for prop in self.__all_props}
                        self.__resolved_hierarchy.update(base.__resolved_hierarchy)

            # property path -> kind
            self.__property_by_path = ComplexKind.resolve_properties(self, model)

            self.__synthetic_props = [p for p in self.__property_by_path if p.prop.synthetic]

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ComplexKind):
            return (
                self.fqn == other.fqn
                and self.properties == other.properties
                and self.bases == other.bases
                and self.allow_unknown_props == other.allow_unknown_props
                and self.successor_kinds == other.successor_kinds
                and self.aggregate_root == other.aggregate_root
                and self.metadata == other.metadata
            )
        else:
            return False

    def __contains__(self, name: str) -> bool:
        return name in self.__prop_by_name

    def __getitem__(self, name: str) -> Property:
        return self.__prop_by_name[name]

    def property_kind_of(self, name: str, or_else: Kind) -> Kind:
        maybe = self.__resolved_kinds.get(name)
        return maybe[1] if maybe else or_else

    def property_with_kind_of(self, name: str) -> Optional[Tuple[Property, Kind]]:
        return self.__resolved_kinds.get(name)

    def property_with_kinds(self) -> Iterable[Tuple[Property, Kind]]:
        return self.__resolved_kinds.values()

    def is_root(self) -> bool:
        return not self.bases or (len(self.bases) == 1 and self.bases[0] == self.fqn)

    def kind_hierarchy(self) -> Set[str]:
        return self.__resolved_hierarchy

    def resolved_properties(self) -> List[ResolvedProperty]:
        return self.__property_by_path

    def resolved_bases(self) -> Dict[str, ComplexKind]:
        return self.__resolved_bases

    def all_props(self) -> List[Property]:
        return self.__all_props

    def all_props_with_kind(self) -> List[Tuple[Property, Kind]]:
        return [(a, self.property_kind_of(a.name, any_kind)) for a in self.__all_props]

    def synthetic_props(self) -> List[ResolvedProperty]:
        return self.__synthetic_props

    def transitive_complex_types(self, with_bases: bool = True, with_properties: bool = True) -> List[ComplexKind]:
        result: Dict[str, ComplexKind] = {}
        visited: Set[str] = set()

        def add_bases(ck: ComplexKind) -> None:
            for fqn, base in ck.resolved_bases().items():
                result[fqn] = base

        def add_props(ck: ComplexKind) -> None:
            for prop in ck.resolved_properties():
                if isinstance(prop.kind, ComplexKind):
                    result[prop.kind.fqn] = prop.kind
                    in_hierarchy(prop.kind)

        def in_hierarchy(ck: ComplexKind) -> None:
            if ck.fqn in visited:
                return
            visited.add(ck.fqn)
            result[ck.fqn] = ck
            if with_bases:
                add_bases(ck)
            if with_properties:
                add_props(ck)
            for base in ck.resolved_bases().values():
                if not base.is_root():
                    in_hierarchy(base)

        in_hierarchy(self)
        return list(result.values())

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        if isinstance(obj, dict):
            coerced = self.coerce_if_required(obj, **kwargs)
            for name, value in (coerced or obj).items():
                known = self.__resolved_kinds.get(name, None)
                if known:
                    prop, kind = known
                    if prop.synthetic:
                        # synthetic properties are computed and will not be maintained. Ignore.
                        pass
                    elif value is None:
                        if prop.required:
                            raise AttributeError(f"Required property {prop.name} is undefined!")
                    else:
                        try:
                            kind.check_valid(value, **kwargs)
                        except AttributeError as at:
                            raise AttributeError(
                                f"Kind:{self.fqn} Property:{name} is not valid: {at}: {json.dumps(obj)}"
                            ) from at
                elif name == "kind":
                    pass
                elif not self.allow_unknown_props:
                    raise AttributeError(f"Kind:{self.fqn} Property:{name} is not defined in model!")
            if not kwargs.get("ignore_missing"):
                for prop in self.__all_props:
                    if prop.required and prop.name not in obj:
                        raise AttributeError(
                            f"Kind:{self.fqn} Property:{prop.name} is required and missing in {json.dumps(obj)}"
                        )
            return coerced
        else:
            raise AttributeError("Kind:{self.fqn} expected a complex type but got this: {obj}")

    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[Json]:
        if isinstance(value, dict):
            result: Json = {}
            has_coerced = False
            for n, v in value.items():
                known = self.__resolved_kinds.get(n, None)
                if known:
                    prop, kind = known
                    if prop.synthetic:
                        # synthetic properties are computed and will not be maintained. Ignore.
                        pass
                    elif v is None:
                        result[n] = None
                    else:
                        coerced = kind.coerce_if_required(v, **kwargs)
                        has_coerced |= coerced is not None
                        result[n] = coerced if coerced is not None else v
                else:
                    result[n] = v
            return result if has_coerced else None
        else:
            return None

    def create_yaml(self, elem: JsonElement, initial_level: int = 0, overrides: Optional[Json] = None) -> str:
        def safe_string(s: str, indent: int, default_style: Optional[str] = None) -> str:
            # 2 spaces per indent
            yml = yaml.dump(s, indent=0, allow_unicode=True, width=sys.maxsize, default_style=default_style)
            yml = remove_suffix(yml, "\n...\n")
            if default_style == "|":
                yml = textwrap.indent(yml, " " * (max(0, indent - 1) * 2)).lstrip()
            else:
                yml = yml.strip()
            return yml

        def override_note(overrides: Optional[Json], prop_name: str) -> Optional[str]:
            if overrides:
                override = overrides.get(prop_name)
                # leaf node of the override tree
                if isinstance(override, (str, list, int, float, bool)):
                    note = (
                        "Warning: the current value is being ignored because there "
                        "is an active override in place. "
                        f"The override value is: {override}"
                    )
                    return note

            return None

        def walk_element(
            e: JsonElement, kind: Kind, indent: int, cr_on_object: bool = True, overrides: Optional[Json] = None
        ) -> str:
            if isinstance(e, dict):
                result = "\n" if cr_on_object else ""
                prepend = "  " * indent
                for prop, value in e.items():
                    description = None
                    sub: Kind = AnyKind()
                    if isinstance(kind, ComplexKind):
                        maybe_prop = kind.property_with_kind_of(prop)
                        if maybe_prop:
                            description = maybe_prop[0].description
                            sub = maybe_prop[1]
                            # if we're in a complex kind, add the override note
                            if note := override_note(overrides, maybe_prop[0].name):
                                description = f"{description}\n{note}" if description else note
                    elif isinstance(kind, DictionaryKind):
                        sub = kind.value_kind
                        # if we're in a dict, add the override note
                        if note := override_note(overrides, prop):
                            description = f"{description}\n{note}" if description else note

                    override_child = None
                    # check if there are overrides to be passed down the tree
                    if overrides:
                        override_child = overrides.get(prop) or {}
                        if not isinstance(override_child, dict):
                            # overrides can't be traversed further
                            override_child = None
                    str_value = walk_element(value, sub, indent + 1, overrides=override_child)
                    if description:
                        for line in description.splitlines():
                            result += f"{prepend}# {line}\n"
                    maybe_space = "" if str_value.startswith("\n") else " "
                    safe_prop = safe_string(prop, indent)
                    result += f"{prepend}{safe_prop}:{maybe_space}{str_value}\n"
                return remove_suffix(result, "\n")
            elif isinstance(e, list) and e:
                prepend = "  " * indent + "-"
                sub = kind.inner if isinstance(kind, ArrayKind) else kind
                result = "\n"
                for item in e:
                    item_str = walk_element(item, sub, indent + 1, False).lstrip()
                    result += f"{prepend} {item_str}\n"
                return remove_suffix(result, "\n")
            elif isinstance(e, list):
                return "[]"
            elif is_env_var_string(e):
                return str(e)
            elif isinstance(e, str):
                # if the string contains a newline, try a literal block, else a quoted string
                return safe_string(e, indent, "|" if ("\n" in e and len(e) > 1) else "'")
            elif e is None:
                return "null"
            elif e is True:
                return "true"
            elif e is False:
                return "false"
            else:
                return str(e)

        return walk_element(elem, self, initial_level, overrides=overrides)

    @staticmethod
    def resolve_properties(
        complex_kind: ComplexKind,
        model: Dict[str, Kind],
        from_path: PropertyPath = EmptyPath,
        maybe_visited: Optional[Dict[str, PropertyPath]] = None,
    ) -> List[ResolvedProperty]:
        visited = maybe_visited or {}
        result: List[ResolvedProperty] = []

        def path_for(
            prop: Property, kind: Kind, path: PropertyPath, array: bool = False, add_prop_to_path: bool = True
        ) -> None:
            prop_name = f"{prop.name}[]" if array else prop.name
            # Detect object cycles: remember the path when we have visited this property.
            # More complex cycles can be detected that way - leave it simple for now.
            key = f"{prop_name}:{prop.kind}"
            if key in visited and prop_name in visited[key].path:
                return
            visited[key] = path
            relative = path.child(prop_name) if add_prop_to_path else path
            # make sure the kind is resolved
            kind.resolve(model)
            if isinstance(kind, SimpleKind):
                result.append(ResolvedProperty(relative, prop, kind))
            elif isinstance(kind, ArrayKind):
                if name := relative.last_part:
                    result.append(ResolvedProperty(relative, Property(name, kind.fqn), kind))
                path_for(prop, kind.inner, path, True)
            elif isinstance(kind, DictionaryKind):
                child = relative.child(None)
                if name := relative.last_part:
                    result.append(ResolvedProperty(relative, Property(name, kind.fqn), kind))
                    # Any child path accessing this dictionary will get a property of value kind.
                    value = kind.value_kind
                    result.append(ResolvedProperty(child, Property("any", value.fqn), value))
                path_for(prop, kind.value_kind, child, add_prop_to_path=False)
            elif isinstance(kind, ComplexKind):
                if name := relative.last_part:
                    result.append(ResolvedProperty(relative, Property(name, kind.fqn), kind))
                result.extend(ComplexKind.resolve_properties(kind, model, relative, visited))

        for x in complex_kind.all_props():
            path_for(x, complex_kind.__resolved_kinds[x.name][1], from_path)

        return result


string_kind = StringKind("string")
int32_kind = NumberKind("int32", "int32")
int64_kind = NumberKind("int64", "int64")
float_kind = NumberKind("float", "float")
double_kind = NumberKind("double", "double")
any_kind = AnyKind()
boolean_kind = BooleanKind("boolean")
date_kind = DateKind("date")
datetime_kind = DateTimeKind("datetime")
duration_kind = DurationKind("duration")

# Define synthetic properties in the metadata section (not defined in the model)
# The predefined_properties kind is used to define the properties there.
synthetic_metadata_kinds = tuple(["exported_age"])

predefined_kinds = [
    string_kind,
    int32_kind,
    int64_kind,
    float_kind,
    double_kind,
    any_kind,
    boolean_kind,
    date_kind,
    datetime_kind,
    duration_kind,
    TransformKind("trafo.duration_to_datetime", "duration", "datetime", "duration_to_datetime", reverse_order=True),
    ComplexKind(
        "graph_root",
        [],
        [
            Property("id", "string", False, None, "The id of this node."),
            Property("name", "string", False, None, "The name of this node."),
            Property("tags", "dictionary[string, string]", False, None, "All attached tags of this node."),
        ],
        allow_unknown_props=True,
    ),
    ComplexKind(
        "predefined_properties",
        [],
        [
            Property("kind", "string", False, None, "The kind property of every node."),
            Property("ctime", "datetime", False, None, "datetime when the node has been created."),
            Property("age", "trafo.duration_to_datetime", False, SyntheticProperty(["ctime"])),
            Property("last_update", "trafo.duration_to_datetime", False, SyntheticProperty(["mtime"])),
            Property("last_access", "trafo.duration_to_datetime", False, SyntheticProperty(["atime"])),
            # Resource tags might define expiration time. Translate durations to absolute times.
            Property("expires", "datetime", False, None, "datetime when the node expires."),
            # Account defines the time when the data was collected in the metadata section
            Property("exported_at", "datetime", False, None, "datetime when the account was exported."),
            Property("exported_age", "trafo.duration_to_datetime", False, SyntheticProperty(["exported_at"])),
        ],
        metadata={"dynamic": True},
    ),
]
predefined_kinds_by_name = {k.fqn: k for k in predefined_kinds}

allowed_simple_type_changes: List[Tuple[Optional[str], Optional[str]]] = [
    ("string", "duration"),
    ("string", "date"),
    ("duration", "string"),
    ("date", "string"),
    ("any", None),
    (None, "any"),
]


class Model:
    @staticmethod
    def empty() -> Model:
        return Model({})

    @staticmethod
    def from_kinds(kinds: List[Kind]) -> Model:
        all_kinds = kinds + predefined_kinds
        kind_dict = {kind.fqn: kind for kind in all_kinds}
        for kind in all_kinds:
            kind.resolve(kind_dict)
        return Model(kind_dict)

    def __init__(self, kinds: Dict[str, Kind]):
        self.kinds = kinds
        self.__property_kind_by_path: List[ResolvedProperty] = list(
            # several complex kinds might have the same property
            # reduce the list by hash over the path.
            {
                r.path: r
                for c in kinds.values()
                if isinstance(c, ComplexKind) and c.aggregate_root
                for r in c.resolved_properties()
            }.values()
        )

    def __contains__(self, name_or_object: Union[str, Json]) -> bool:
        if isinstance(name_or_object, str):
            return name_or_object in self.kinds
        elif isinstance(name_or_object, dict) and "kind" in name_or_object:
            return name_or_object["kind"] is self.kinds
        else:
            return False

    def __getitem__(self, name_or_object: Union[str, Json]) -> Kind:
        if isinstance(name_or_object, str):
            return self.kinds[name_or_object]
        elif isinstance(name_or_object, dict) and "kind" in name_or_object:
            return self.kinds[name_or_object["kind"]]
        else:
            raise KeyError(f"Expected string or json with a 'kind' property as key but got: {name_or_object}")

    def __len__(self) -> int:
        return len(self.kinds)

    def get(self, name_or_object: Union[str, Json]) -> Optional[Kind]:
        if isinstance(name_or_object, str):
            return self.kinds.get(name_or_object)
        elif isinstance(name_or_object, dict) and "kind" in name_or_object:
            return self.kinds.get(name_or_object["kind"])
        else:
            return None

    def complex_kinds(self) -> List[ComplexKind]:
        return [k for k in self.kinds.values() if isinstance(k, ComplexKind)]

    def property_by_path(self, path_: str) -> ResolvedProperty:
        path = PropertyPath.from_path(path_)
        found: Optional[ResolvedProperty] = first(lambda prop: prop.path.same_as(path), self.__property_kind_by_path)
        # if the path is not known according to known model: it could be anything.
        return found if found else ResolvedProperty(path, Property.any_prop(), AnyKind())

    def kind_by_path(self, path_: str) -> Kind:
        return self.property_by_path(path_).kind

    def coerce(self, js: Json) -> Json:
        try:
            kind: Kind = self[js["kind"]]
            return kind.coerce(js)  # type: ignore
        except KeyError as ex:
            raise AttributeError(
                f'No kind definition found for {js["kind"]}' if "kind" in js else f"No attribute kind found in {js}"
            ) from ex

    def check_valid(self, js: Json, **kwargs: bool) -> ValidationResult:
        try:
            kind: Kind = self[js["kind"]]
            return kind.check_valid(js, **kwargs)
        except KeyError as ex:
            raise AttributeError(
                f'No kind definition found for {js["kind"]}' if "kind" in js else f"No attribute kind found in {js}"
            ) from ex

    @lru_cache
    def predefined_synthetic_props(self, allowed: Tuple[str, ...]) -> List[ResolvedProperty]:
        predefined = self.get("predefined_properties")
        if isinstance(predefined, ComplexKind):
            return [p for p in predefined.synthetic_props() if p.prop.name in allowed]
        else:
            return []

    def graph(self) -> MultiDiGraph:
        graph = MultiDiGraph()

        def handle_complex(cx: ComplexKind) -> None:
            # do not handle the same complex kind more than once
            if cx.fqn in graph and "data" in graph.nodes[cx.fqn]:
                return

            graph.add_node(cx.fqn, data=cx)
            # inheritance
            if not cx.is_root():
                for base in cx.bases:
                    graph.add_edge(cx.fqn, base, f"inheritance_{cx.fqn}_{base}", type="inheritance")

            # properties
            for _, prop_kind in cx.property_with_kinds():
                for cpl in prop_kind.nested_complex_kinds():
                    graph.add_edge(cx.fqn, cpl.fqn, f"property_{cx.fqn}_{cpl.fqn}", type="property")
                    handle_complex(cpl)

            # dependency
            for name, successors in cx.successor_kinds.items():
                for successor in successors or []:
                    graph.add_edge(
                        cx.fqn, successor, f"successor_{cx.fqn}_{successor}_{name}", type="successor", edge_type=name
                    )

        for kind in self.kinds.values():
            if isinstance(kind, ComplexKind):
                handle_complex(kind)
            elif isinstance(kind, ArrayKind) and isinstance(kind.inner, ComplexKind):
                handle_complex(kind.inner)
            elif isinstance(kind, DictionaryKind) and isinstance(kind.value_kind, ComplexKind):
                handle_complex(kind.value_kind)

        return graph

    def update_kinds(self, kinds: List[Kind], check_overlap: bool = True) -> Model:
        # Create a list of kinds that have changed to the existing model
        to_update = []

        def merge_metadata(update: ComplexKind, existing: ComplexKind) -> None:
            """
            Merge metadata of complex kinds and their properties.
            ATM: we only merge the len property of string kinds.
            """
            existing_props: Dict[str, Property] = {prop.name: prop for prop in existing.properties}
            for prop in update.properties:
                existing_prop = existing_props.get(prop.name)
                meta = prop.metadata or {}
                # take the maximum of all lengths ever seen
                if existing_prop is not None and existing_prop.meta("len", int) is not None:
                    meta["len"] = max(prop.meta_get("len", int, 0), existing_prop.meta_get("len", int, 0))
                    prop.metadata = meta

        for elem in kinds:
            existing_elem = self.kinds.get(elem.fqn)
            # merge old and new metadata for complex kinds and their properties
            if isinstance(elem, ComplexKind) and isinstance(existing_elem, ComplexKind):
                merge_metadata(elem, existing_elem)
            if elem.fqn not in predefined_kinds_by_name and (not existing_elem or existing_elem != elem):
                to_update.append(elem)

        # Short circuit, if there are no changes
        if not to_update:
            return self

        def update_is_valid(from_kind: Kind, to_kind: Kind) -> None:
            # Allowed changes: The update
            # - does not change it's type (e.g. going from SimpleKind to ComplexKind)
            if type(from_kind) != type(to_kind):  # pylint: disable=unidiomatic-typecheck
                raise AttributeError(f"Update {from_kind.fqn} changes an existing property type {from_kind.fqn}")

        # resolve and build dict
        updates = {elem.fqn: elem for elem in to_update}
        updated = {**self.kinds, **updates}
        for elem in to_update:
            elem.resolve(updated)

        # check if updating existing kinds is allowed
        for name in self.kinds.keys() & updates.keys():
            update_is_valid(self.kinds[name], updates[name])

        if check_overlap:
            check_overlap_for([to_js(a) for a in updated.values()])

        return Model(updated)

    def flat_kinds(self) -> List[Kind]:
        """
        Returns a list of all kinds. The hierarchy of complex kinds is flattened:
        - all properties of all base kinds.
        - all metadata merged in hierarchy.
        - all successor kinds combined in hierarchy.
        """
        cpl: Dict[str, ComplexKind] = {kind.fqn: kind for kind in self.complex_kinds()}

        def all_props(kind: ComplexKind) -> Dict[str, Property]:
            props_by_name = {}
            for props in [all_props(cpl[fqn]) for fqn in kind.bases] + [{p.name: p for p in kind.properties}]:
                for key, value in props.items():
                    props_by_name[key] = value
            return props_by_name

        def all_metadata(kind: ComplexKind) -> Dict[str, Any]:
            metadata = {}
            for meta in [all_metadata(cpl[fqn]) for fqn in kind.bases] + [kind.metadata]:
                for key, value in meta.items():
                    metadata[key] = value
            return metadata

        def all_successor_kinds(kind: ComplexKind) -> Dict[EdgeType, List[str]]:
            successor_kinds = kind.successor_kinds.copy()
            for base in kind.bases:
                for edge_type, succ in all_successor_kinds(cpl[base]).items():
                    successor_kinds[edge_type] = successor_kinds.get(edge_type, []) + succ
            return successor_kinds

        result: List[Kind] = []
        for kind in self.kinds.values():
            if isinstance(kind, ComplexKind):
                result.append(
                    ComplexKind(
                        fqn=kind.fqn,
                        bases=kind.bases,
                        properties=list(all_props(kind).values()),
                        allow_unknown_props=kind.allow_unknown_props,
                        successor_kinds=all_successor_kinds(kind),
                        aggregate_root=kind.aggregate_root,
                        metadata=all_metadata(kind),
                    )
                )
            else:
                result.append(kind)
        return result


class UsageMetricValues(NamedTuple):
    min: float
    avg: float
    max: float


@frozen
class UsageDatapoint:
    """
    A single datapoint of resource usage.

    id: `str`
        Identifier of the resource as named by the cloud.
    at: `int`
        Timestamp of the datapoint in seconds since epoch.
        Name of the metric.
    change_id: `str`
        Id of the import
    v: `Dict[str, List[float]]]`
        Dictionary of metric names to lists of values. The values are `min`, `avg` and `max`.

    """

    id: str
    at: int
    change_id: str
    v: Dict[str, UsageMetricValues]


# register serializer for this class
set_deserializer(Kind.from_json, Kind)
set_serializer(SimpleKind.to_json, SimpleKind)



import json
import shutil
from copy import deepcopy
from datetime import datetime, timezone

import pytest
import pytz
from aiostream import stream

from resotocore.util import (
    AccessJson,
    force_gen,
    uuid_str,
    value_in_path,
    value_in_path_get,
    set_value_in_path,
    rnd_str,
    del_value_in_path,
    deep_merge,
    partition_by,
    utc_str,
    parse_utc,
)


def not_in_path(name: str, *other: str) -> bool:
    for n in [name, *other]:
        if shutil.which(n) is None:
            return True
    return False


def test_partition_by() -> None:
    even, odd = partition_by(lambda x: x % 2 == 0, range(10))
    assert even == [0, 2, 4, 6, 8]
    assert odd == [1, 3, 5, 7, 9]


def test_access_json() -> None:
    js = {"a": "a", "b": {"c": "c", "d": {"e": "e", "f": [0, 1, 2, 3, 4]}}}
    access = AccessJson(js, "null", self_name="this")

    assert access.a == "a"
    assert access.b.d.f[2] == 2
    assert access.this == js
    assert str(access.b.d.f[99]) is "null"
    assert str(access.b.d.f["test"]) is "null"
    assert str(access.foo.bla.bar[23].now) is "null"
    assert json.dumps(access.b.d, sort_keys=True) == '{"e": "e", "f": [0, 1, 2, 3, 4]}'

    assert access["a"] == "a"
    assert access["b"]["d"]["f"][2] == 2
    assert str(access["foo"]["bla"]["bar"][23]["now"]) is "null"
    assert json.dumps(access["b"]["d"], sort_keys=True) == '{"e": "e", "f": [0, 1, 2, 3, 4]}'

    assert [a for a in access] == ["a", "b"]
    assert [a for a in access.doesnt.exist] == []
    assert [a for a in access.b.d.items()] == [("e", "e"), ("f", [0, 1, 2, 3, 4])]


def test_uuid() -> None:
    assert uuid_str("foo") == uuid_str("foo")
    assert uuid_str("foo") != uuid_str("bla")
    assert uuid_str() != uuid_str()


def test_random_str() -> None:
    assert rnd_str() != rnd_str()


def test_value_in_path() -> None:
    js = {"foo": {"bla": {"test": 123}}, "b": [{"a": 1, "b": [1, 2, 3]}, {"a": 2, "b": [1, 2, 3]}]}
    assert value_in_path(js, ["foo", "bla", "test"]) == 123
    assert value_in_path_get(js, ["foo", "bla", "test"], "foo") == "foo"  # expected string got int -> default value
    assert value_in_path(js, ["foo", "bla", "test", "bar"]) is None
    assert value_in_path_get(js, ["foo", "bla", "test", "bar"], 123) == 123
    assert value_in_path(js, ["foo", "bla", "bar"]) is None
    assert value_in_path_get(js, ["foo", "bla", "bar"], "foo") == "foo"


def test_set_value_in_path() -> None:
    js = {"foo": {"bla": {"test": 123}}}
    res = set_value_in_path(124, ["foo", "bla", "test"], deepcopy(js))
    assert res == {"foo": {"bla": {"test": 124}}}
    res = set_value_in_path(124, ["foo", "bla", "blubber"], deepcopy(js))
    assert res == {"foo": {"bla": {"test": 123, "blubber": 124}}}
    res = set_value_in_path(js, ["foo", "bla", "test"])
    assert res == {"foo": {"bla": {"test": js}}}
    res = {"a": 1}
    set_value_in_path(23, ["reported"], res)
    assert res == {"a": 1, "reported": 23}


def test_del_value_in_path() -> None:
    js = {"foo": {"bla": {"test": 123}}}
    res = del_value_in_path(deepcopy(js), ["foo", "bla", "test"])
    assert res == {"foo": {"bla": None}}
    js = {"foo": {"bla": {"test": 123}}}
    res = del_value_in_path(deepcopy(js), ["foo", "bla", "bar"])
    assert res == {"foo": {"bla": {"test": 123}}}
    js = {"foo": {"bla": {"test": 123}}}
    res = del_value_in_path(deepcopy(js), ["foo", "bla"])
    assert res == {"foo": None}
    res = del_value_in_path(deepcopy(js), ["foo"])
    assert res == {}


@pytest.mark.asyncio
async def test_async_gen() -> None:
    async with stream.empty().stream() as empty:
        async for _ in await force_gen(empty):
            pass

    with pytest.raises(Exception):
        async with stream.throw(Exception(";)")).stream() as err:
            async for _ in await force_gen(err):
                pass

    async with stream.iterate(range(0, 100)).stream() as elems:
        assert [x async for x in await force_gen(elems)] == list(range(0, 100))


def test_deep_merge() -> None:
    l = {"a": {"b": 1, "d": 2}, "d": 2, "e": 4}
    r = {"a": {"c": 1, "d": 3}, "d": 1}
    assert deep_merge(l, r) == {"a": {"b": 1, "c": 1, "d": 3}, "d": 1, "e": 4}

    a = {"a": {"foo": {"first": "first", "last": "laaaast"}}, "b": {"bar": 123}, "c": [6, 7]}
    b = {"a": {"foo": {"last": "last"}}, "b": {"baz": 456}, "c": [8, 9]}
    assert deep_merge(a, b) == {
        "a": {"foo": {"first": "first", "last": "last"}},
        "b": {"bar": 123, "baz": 456},
        "c": [8, 9],
    }


def test_utc() -> None:
    # datetime in utc
    assert utc_str(datetime(2020, 1, 1, 0, 0, 0, 0, tzinfo=timezone.utc)) == "2020-01-01T00:00:00Z"
    # no timezone: assume utc
    assert utc_str(datetime(2020, 1, 1, 0, 0, 0, 0)) == "2020-01-01T00:00:00Z"
    # different timezone: convert to utc
    assert utc_str(datetime(2020, 1, 1, 0, 0, 0, 0, tzinfo=pytz.timezone("CET"))) == "2019-12-31T23:00:00Z"
    # gmt same timezone as utc
    assert utc_str(datetime(2020, 1, 1, 0, 0, 0, 0, tzinfo=pytz.timezone("GMT"))) == "2020-01-01T00:00:00Z"

    # valid utc
    assert parse_utc("2020-01-01T00:00:00Z") == datetime(2020, 1, 1, 0, 0, 0, 0, tzinfo=timezone.utc)
    # no timezone: assume utc
    assert parse_utc("2020-01-01T00:00:00") == datetime(2020, 1, 1, 0, 0, 0, 0, tzinfo=timezone.utc)
    # different timezone: convert to utc
    assert parse_utc("2019-12-31T23:00:00-01:00") == datetime(2020, 1, 1, 0, 0, 0, 0, tzinfo=timezone.utc)
    assert parse_utc("2019-12-31T22:00:00-02:00") == datetime(2020, 1, 1, 0, 0, 0, 0, tzinfo=timezone.utc)
    assert parse_utc("2020-01-01T00:00:00+00:00") == datetime(2020, 1, 1, 0, 0, 0, 0, tzinfo=timezone.utc)
    # other formats
    assert parse_utc("20200101T000000Z") == datetime(2020, 1, 1, 0, 0, 0, 0, tzinfo=timezone.utc)
    assert parse_utc("20200101000000") == datetime(2020, 1, 1, 0, 0, 0, 0, tzinfo=timezone.utc)
    assert parse_utc("20200101") == datetime(2020, 1, 1, 0, 0, 0, 0, tzinfo=timezone.utc)
    assert parse_utc("2020-01-01") == datetime(2020, 1, 1, 0, 0, 0, 0, tzinfo=timezone.utc)



import logging
from typing import Optional, MutableMapping, Union, Tuple

from arango import HTTPClient
from arango.response import Response
from arango.typings import Headers
from requests import Session
from requests.adapters import HTTPAdapter, Retry
from requests_toolbelt import MultipartEncoder

log = logging.getLogger(__name__)


class ArangoHTTPClient(HTTPClient):
    def __init__(self, timeout: int, verify: Union[str, bool, None]):
        log.info(f"Create ArangoHTTPClient with timeout={timeout} and verify={verify}")
        self.timeout = timeout
        self.verify = verify

    def create_session(self, host: str) -> Session:
        retry_strategy = Retry(
            total=3,
            backoff_factor=1,
            status_forcelist=[429, 500, 502, 503, 504],
            allowed_methods=["HEAD", "GET", "OPTIONS"],
        )
        http_adapter = HTTPAdapter(max_retries=retry_strategy)
        session = Session()
        session.mount("https://", http_adapter)
        session.mount("http://", http_adapter)
        session.verify = self.verify
        return session

    def send_request(
        self,
        session: Session,
        method: str,
        url: str,
        headers: Optional[Headers] = None,
        params: Optional[MutableMapping[str, str]] = None,
        data: Union[str, MultipartEncoder, None] = None,
        auth: Optional[Tuple[str, str]] = None,
    ) -> Response:
        response = session.request(method, url, params, data, headers, auth=auth, timeout=self.timeout)
        return Response(method, response.url, response.headers, response.status_code, response.reason, response.text)



from hypothesis.strategies import (
    composite,
    SearchStrategy,
    lists,
    sampled_from,
    builds,
    just,
    integers,
    booleans,
    tuples,
)

from resotocore.model.graph_access import EdgeTypes, Direction
from tests.resotocore.hypothesis_extension import Drawer, optional, UD, any_string, any_datetime
from resotocore.query.model import (
    IsTerm,
    Predicate,
    Term,
    Part,
    Query,
    CombinedTerm,
    Navigation,
    WithClause,
    WithClauseFilter,
    SortOrder,
    Sort,
    MergeTerm,
    MergeQuery,
    AllTerm,
    Aggregate,
    AggregateVariableName,
    AggregateVariableCombined,
    AggregateVariable,
    AggregateFunction,
    Limit,
    WithUsage,
)


query_property = sampled_from(["reported.name", "reported.cpu_count"])
kind = sampled_from(["bucket", "volume", "certificate", "cloud", "database", "endpoint"])
query_operations = sampled_from(["==", ">=", "<=", ">", "<"])
query_values = sampled_from(["test", 23, True, False, None])
combine_term = sampled_from(["and", "or"])
edge_direction = sampled_from(list(Direction.all))
edge_type = sampled_from(list(EdgeTypes.all))
sort_order = sampled_from([SortOrder.Asc, SortOrder.Desc])
aggregate_functions = sampled_from(["sum", "count", "min", "max", "avg"])
is_term = builds(IsTerm, lists(kind, min_size=1, max_size=2))
predicate_term = builds(Predicate, query_property, query_operations, query_values, just({}))
leaf_term: SearchStrategy[Term] = is_term | predicate_term
limit_gen = builds(Limit, integers(min_value=0), integers(min_value=1))


@composite
def composite_term(ud: UD) -> CombinedTerm:
    d = Drawer(ud)
    trm = leaf_term | composite_term()
    return CombinedTerm(d.draw(trm), d.draw(combine_term), d.draw(trm))


term: SearchStrategy[Term] = leaf_term | composite_term()
sort = builds(Sort, query_property, sort_order)


@composite
def merge_term(ud: UD) -> MergeTerm:
    d = Drawer(ud)
    pre = d.draw(term)
    queries = d.draw(lists(merge_query, min_size=1, max_size=1))
    pst = d.optional(term)
    return MergeTerm(pre, queries, pst)


@composite
def navigation(ud: UD) -> Navigation:
    d = Drawer(ud)
    start = d.draw(integers(min_value=0, max_value=1))
    length = d.draw(integers(min_value=0, max_value=100))
    ed = d.draw(edge_type)
    direction = d.draw(edge_direction)
    return Navigation(start, length + start, [ed], direction)


@composite
def with_clause(ud: UD) -> WithClause:
    d = Drawer(ud)
    op = d.draw(query_operations)
    num = d.draw(integers(min_value=0))
    nav = d.draw(navigation())
    trm = d.optional(term)
    wc = d.optional(with_clause())
    return WithClause(WithClauseFilter(op, num), nav, trm, wc)


@composite
def with_usage(ud: UD) -> WithUsage:
    d = Drawer(ud)
    start = d.draw(any_datetime)
    end = d.optional(any_datetime)
    metrics = d.draw(lists(any_string, min_size=1, max_size=3))
    return WithUsage(start, end, metrics)


part = builds(
    Part,
    term | merge_term(),
    optional(any_string),
    with_clause(),
    with_usage(),
    lists(sort, min_size=0, max_size=3),
    optional(limit_gen),
    navigation(),
)

only_filter_part = builds(
    Part, term, just(None), just(None), just(None), lists(sort, min_size=0, max_size=1), optional(limit_gen)
)


@composite
def merge_query_query(ud: UD) -> Query:
    d = Drawer(ud)
    nav = d.draw(navigation())
    trm = d.draw(term)
    # merge query need to start with navigation part without additional props
    parts = [Part(trm), Part(AllTerm(), navigation=nav)]
    return Query(parts)


merge_query = builds(MergeQuery, any_string, merge_query_query(), booleans())


@composite
def aggregate_variable_combined(ud: UD) -> AggregateVariableCombined:
    d = Drawer(ud)
    return AggregateVariableCombined([d.draw(any_string), d.draw(aggregate_variable_name), d.draw(any_string)])


aggregate_variable_name = builds(AggregateVariableName, any_string)
aggregate_variable = builds(AggregateVariable, aggregate_variable_name | aggregate_variable_combined())
aggregate_op = sampled_from(["+", "-", "*", "/"])
aggregate_function = builds(
    AggregateFunction, aggregate_functions, any_string | integers(), lists(tuples(aggregate_op, integers())), any_string
)
aggregate = builds(Aggregate, lists(aggregate_variable, min_size=1), lists(aggregate_function, min_size=1))
query = builds(
    Query,
    # test a more complex query with multiple parts and navigation or one simple part without navigation
    lists(part, min_size=1, max_size=3) | lists(only_filter_part, min_size=1, max_size=1),
    just({}),
    optional(aggregate),
)



from typing import Any, Optional, Callable, Dict, Sequence, Union, Mapping, Literal

from jsons import set_deserializer

# Ideally we would be able to define it like this:
# JsonElement = Union[ Mapping[str, JsonElement], Sequence[JsonElement], str, int, float, bool, None]
# Sadly python does not support recursive types yet, so we try to narrow it to:
JsonElement = Union[
    str,
    int,
    float,
    bool,
    None,
    Mapping[str, Any],
    Sequence[Union[Mapping[str, Any], Sequence[Any], str, int, float, bool, None]],
]
JsonArray = Sequence[JsonElement]
Json = Dict[str, Any]


ValidationResult = Optional[Any]
ValidationFn = Callable[[Any], ValidationResult]

EdgeType = Literal["default", "delete"]


# make sure jsons does not do something clever, when a json element needs to be parsed
# omitting this deserializer, will read a string into a list of characters etc.
def parse_js_element(elem: JsonElement, _: type = object, **kwargs: object) -> JsonElement:
    return elem


set_deserializer(parse_js_element, [str, int, float, bool, None])



from textwrap import dedent
from typing import List, cast, Dict
import os

import pytest
from pytest import fixture

from resotocore.analytics import InMemoryEventSender
from resotocore.config import ConfigHandler, ConfigEntity, ConfigValidation, ConfigOverride
from resotocore.config.config_handler_service import ConfigHandlerService
from resotocore.ids import ConfigId
from resotocore.message_bus import CoreMessage, Event, Message
from tests.resotocore.message_bus_test import wait_for_message
from resotocore.model.model import Kind, ComplexKind, Property
from resotocore.model.typed_model import to_js, from_js
from resotocore.types import Json
from types import SimpleNamespace


@fixture
def config_model() -> List[Kind]:
    return [
        ComplexKind(
            "sub_section",
            [],
            [
                Property("num", "int32", description="Some arbitrary number."),
                Property("str", "string", description="Some arbitrary string."),
            ],
        ),
        ComplexKind(
            "section",
            [],
            [
                Property("some_number", "int32", required=True, description="Some number.\nAnd some description."),
                Property(
                    "some_number_env_var",
                    "int32",
                    required=False,
                    description="Some env var substitution.\nAnd some description.",
                ),
                Property("some_string", "string", required=True, description="Some string.\nAnd some description."),
                Property("some_sub", "sub_section", required=True, description="Some sub.\nAnd some description."),
            ],
        ),
    ]


@pytest.mark.asyncio
async def test_config(config_handler: ConfigHandler) -> None:
    # list is empty on start
    assert [a async for a in config_handler.list_config_ids()] == []

    config_id = ConfigId("test")
    # add one entry
    entity = ConfigEntity(config_id, {"test": True})
    assert await config_handler.put_config(entity) == entity

    # get one entry
    assert await config_handler.get_config(config_id) == entity

    # patch the config
    assert await config_handler.patch_config(ConfigEntity(config_id, {"rest": False})) == ConfigEntity(
        config_id, {"test": True, "rest": False}
    )

    # list all configs
    assert [a async for a in config_handler.list_config_ids()] == ["test"]

    # copy the config
    copied = await config_handler.copy_config(config_id, ConfigId("test2"))
    # non-existing config is not copied
    assert await config_handler.copy_config(ConfigId("test3"), ConfigId("test4")) is None
    # copying to existing config is not allowed and raises an exception
    with pytest.raises(Exception):
        await config_handler.copy_config(ConfigId("test"), ConfigId("test"))
    assert copied == ConfigEntity(ConfigId("test2"), {"test": True, "rest": False})
    assert {a async for a in config_handler.list_config_ids()} == set(["test", "test2"])

    # delete the config
    await config_handler.delete_config(config_id)
    await config_handler.delete_config(ConfigId("test2"))

    # list all configs
    assert [a async for a in config_handler.list_config_ids()] == []


@pytest.mark.asyncio
async def test_config_change_event(config_handler: ConfigHandlerService) -> None:
    sender: InMemoryEventSender = config_handler.event_sender  # type: ignore # in-memory sender
    # list of events is empty on start
    assert sender.events == []

    config_id = ConfigId("test")
    # add one entry
    entity = ConfigEntity(config_id, {"test": True})
    assert await config_handler.put_config(entity) == entity
    assert len(sender.events) == 1

    # patch the config
    assert await config_handler.patch_config(ConfigEntity(config_id, {"rest": False})) == ConfigEntity(
        config_id, {"test": True, "rest": False}
    )
    assert len(sender.events) == 2

    # delete the config
    await config_handler.delete_config(config_id)
    assert len(sender.events) == 3


@pytest.mark.asyncio
async def test_dry_run(config_handler: ConfigHandlerService) -> None:
    config_id = ConfigId("test_dry_run")
    entity = ConfigEntity(config_id, {"test": True})

    # put dry run -> config is not stored
    assert await config_handler.put_config(entity, dry_run=True) == entity
    assert await config_handler.get_config(config_id) is None

    # patch dry run -> config is not stored
    assert await config_handler.patch_config(entity, dry_run=True) == entity
    assert await config_handler.get_config(config_id) is None


@pytest.mark.asyncio
async def test_config_change_analytics(config_handler: ConfigHandler) -> None:
    config_id = ConfigId("test")
    worker_config_1 = {
        "resotoworker": {
            "collector": ["aws", "k8s", "example", "digitalocean", "gcp"],
            "aws": {"access_key_id": None, "secret_access_key": None, "profiles": ["list", "of", "profiles"]},
            "digitalocean": {"api_tokens": ["123abc"]},
            "gcp": {"service_account": [""]},
            "k8s": {"config_files": ["/path/to/some/file"]},
        }
    }
    entity = ConfigEntity(config_id, worker_config_1)
    analytics = entity.analytics()

    assert analytics["how_many_providers"] == 4
    assert analytics["aws"]
    assert analytics["k8s"]
    assert analytics["gcp"]
    assert analytics["digitalocean"]
    assert analytics["aws_use_profiles"]
    assert not analytics["aws_use_role"]
    assert analytics["do_use_config"]
    assert analytics["gcp_use_auto_discovery"]
    assert analytics["k8s_use_kubeconfig"]

    worker_config_2 = {
        "resotoworker": {
            "collector": ["aws", "k8s", "example", "digitalocean", "gcp"],
            "aws": {
                "access_key_id": "abc",
                "secret_access_key": "123",
            },
            "gcp": {"service_account": ["some service account json file"]},
            "k8s": {
                "configs": [
                    {
                        "name": "dev",
                        "certificate_authority_data": "xyz",
                        "server": "https://k8s-cluster-server.example.com",
                        "token": "some token",
                    }
                ]
            },
        }
    }
    entity = ConfigEntity(config_id, worker_config_2)
    analytics = entity.analytics()
    assert not analytics["aws_use_profiles"]
    assert analytics["aws_use_access_secret_key"]
    assert not analytics["do_use_config"]
    assert analytics["gcp_use_file"]
    assert analytics["k8s_use_manual"]


@pytest.mark.asyncio
async def test_config_validation(config_handler: ConfigHandler, config_model: List[Kind]) -> None:
    await config_handler.update_configs_model(config_model)
    valid_config = {"section": {"some_number": 32, "some_string": "test", "some_sub": {"num": 32}}}

    # define the model
    await config_handler.put_config_validation(ConfigValidation("test", True))

    # list all available models
    assert [a async for a in config_handler.list_config_validation_ids()] == ["test"]

    # get the model
    model: ConfigValidation = await config_handler.get_config_validation("test")  # type: ignore
    assert model.external_validation is True

    # check the config against the model
    invalid_config = {"section": {"some_number": "no number"}}
    invalid_config_id = ConfigId("invalid_config")
    with pytest.raises(AttributeError) as reason:
        await config_handler.put_config(ConfigEntity(invalid_config_id, invalid_config))
    assert "some_number is not valid: Expected type int32 but got str" in str(reason.value)

    # External validation turned on: config with name "invalid_config" is rejected by the configured worker
    await config_handler.put_config_validation(ConfigValidation(invalid_config_id, True))
    with pytest.raises(AttributeError) as reason:
        # The config is actually valid, but the external validation will fail
        await config_handler.put_config(ConfigEntity(invalid_config_id, valid_config))
    assert "Error executing task: Invalid Config ;)" in str(reason)

    # If external validation is turned off, the configuration can be updated
    await config_handler.put_config_validation(ConfigValidation(invalid_config_id, False))
    await config_handler.put_config(ConfigEntity(invalid_config_id, valid_config))


@pytest.mark.asyncio
async def test_config_yaml(config_handler: ConfigHandler, config_model: List[Kind]) -> None:
    await config_handler.update_configs_model(config_model)
    config = {
        "some_number": 32,
        "some_number_env_var": "$(SOME_NUMBER)",
        "some_string": "test",
        "some_sub": {"num": 32},
    }
    os.environ["SOME_NUMBER"] = "42"
    expect_comment = dedent(
        """
        section:
          # Some number.
          # And some description.
          some_number: 32
          # Some env var substitution.
          # And some description.
          some_number_env_var: $(SOME_NUMBER)
          # Some string.
          # And some description.
          some_string: 'test'
          # Some sub.
          # And some description.
          some_sub:
            # Some arbitrary number.
            num: 32
        """
    ).strip()
    # no attached model -> '32' is not coerced into 32
    expect_no_comment = dedent(
        """
        another_section:
          some_number: 32
          some_number_env_var: $(SOME_NUMBER)
          some_string: test
          some_sub:
            num: 32
        """
    ).strip()
    # config has section with attached model
    test_config_id = ConfigId("test")
    await config_handler.put_config(ConfigEntity(test_config_id, {"section": config}))
    config_yaml = await config_handler.config_yaml(test_config_id) or ""
    assert expect_comment in config_yaml
    # different section with no attached model
    nomodel_config_id = ConfigId("no_model")
    await config_handler.put_config(ConfigEntity(nomodel_config_id, {"another_section": config}))
    nomodel_config_yaml = await config_handler.config_yaml(nomodel_config_id) or ""
    assert expect_no_comment in nomodel_config_yaml

    expect_json = {
        "some_number": 32,
        "some_number_env_var": 42,
        "some_string": "test",
        "some_sub": {"num": 32},
    }
    # get config returns the latest version
    stored_ce = await config_handler.get_config(test_config_id)
    stored_config = stored_ce.config.get("section") if stored_ce else None
    assert expect_json == stored_config

    override = {"section": {"some_number": 1337}}
    expect_override_comment = dedent(
        """
        section:
          # Some number.
          # And some description.
          # Warning: the current value is being ignored because there is an active override in place. The override value is: 1337
          some_number: 32
          # Some env var substitution.
          # And some description.
          some_number_env_var: $(SOME_NUMBER)
          # Some string.
          # And some description.
          some_string: 'test'
          # Some sub.
          # And some description.
          some_sub:
            # Some arbitrary number.
            num: 32
        """
    ).strip()
    cast(ConfigHandlerService, config_handler).override_service = cast(
        ConfigOverride, SimpleNamespace(get_override=lambda id: override)
    )
    config_with_override_yaml = await config_handler.config_yaml(test_config_id) or ""
    assert expect_override_comment in config_with_override_yaml


@pytest.mark.asyncio
async def test_config_change_emits_event(config_handler: ConfigHandler, all_events: List[Message]) -> None:
    # Put a config
    all_events.clear()
    config_id = ConfigId("foo")
    cfg = await config_handler.put_config(ConfigEntity(config_id, dict(test=1)))
    message = await wait_for_message(all_events, CoreMessage.ConfigUpdated, Event)
    assert message.data["id"] == cfg.id
    assert message.data["revision"] == cfg.revision

    # Patch a config
    all_events.clear()
    cfg = await config_handler.patch_config(ConfigEntity(config_id, dict(foo=2)))
    message = await wait_for_message(all_events, CoreMessage.ConfigUpdated, Event)
    assert message.data["id"] == cfg.id
    assert message.data["revision"] == cfg.revision

    # Delete a config
    all_events.clear()
    await config_handler.delete_config(config_id)
    message = await wait_for_message(all_events, CoreMessage.ConfigDeleted, Event)
    assert message.data["id"] == config_id
    assert "revision" not in message.data


def test_config_entity_roundtrip() -> None:
    entity = ConfigEntity(ConfigId("test"), {"test": 1}, "test")
    again = from_js(to_js(entity), ConfigEntity)
    assert entity == again



from abc import ABC, abstractmethod
from typing import Dict, List, ClassVar, Optional, Type, Set, Any

from attr import define, field

from resotocore.ids import ConfigId, Email, Password
from resotocore.model.typed_model import to_js
from resotocore.types import Json
from resotocore.service import Service
from resotolib.core.model_export import dataclasses_to_resotocore_model

UsersConfigRoot = "resoto_users"
UsersConfigId = ConfigId("resoto.users")


ValidRoles = {"admin", "readwrite", "readonly"}


@define
class ResotoUser:
    kind: ClassVar[str] = "resoto_user"
    fullname: str = field(metadata={"description": "The full name of the user."})
    password_hash: str = field(metadata={"description": "The sha256 hash of the user's password."})
    roles: Set[str] = field(factory=set, metadata={"description": "The roles of the user."})


@define
class ResotoUsersConfig:
    kind: ClassVar[str] = UsersConfigRoot
    users: Dict[str, ResotoUser] = field(factory=lambda: {}, metadata={"description": "A map of email to user data."})

    def json(self) -> Json:
        return {UsersConfigRoot: to_js(self, strip_attr="kind")}


class UserManagement(Service, ABC):
    @abstractmethod
    async def has_users(self) -> bool:
        """
        Indicates if users exist in the system.
        """

    @abstractmethod
    async def create_first_user(self, company: str, fullname: str, email: Email, password: Password) -> ResotoUser:
        """
        Create the first user in the system.
        Precondition: has_users() == False
        :param company: the name of the company
        :param fullname: the full name of the user
        :param email: the email address of the user
        :param password: the password of the user
        :return: the created user
        :throws: AssertionError if there are already users in the system.
        """

    @abstractmethod
    async def login(self, email: Email, password: Password) -> Optional[ResotoUser]:
        """
        Login with the given credentials.
        :param email: the email address of the user
        :param password: the password of the user
        :return: The user if the credentials are valid, None otherwise.
        """

    @abstractmethod
    async def create_user(self, email: Email, fullname: str, password: Password, roles: List[str]) -> ResotoUser:
        """
        Create a new user.
        :param email: the email address of the user
        :param fullname: the full name of the user
        :param password: password of the user
        :param roles: all roles of the user
        :return: the created user
        """

    @abstractmethod
    async def update_user(
        self, email: Email, *, password: Optional[Password] = None, roles: Optional[List[str]] = None
    ) -> ResotoUser:
        """
        Update an existing user.

        :param email: the email address of the user
        :param password: the new password of the user
        :param roles: all roles of the user
        :return: the updated user
        """

    @abstractmethod
    async def delete_user(self, email: Email) -> Optional[ResotoUser]:
        """
        Delete an existing user.
        :param email: the email address of the user
        :return: the deleted user if it existed before calling this method, None otherwise
        """

    @abstractmethod
    async def user(self, email: Email) -> Optional[ResotoUser]:
        """
        Get a user by email.
        :param email: the email address of the user
        :return: the user if it exists, None otherwise
        """

    @abstractmethod
    async def users(self) -> Dict[Email, ResotoUser]:
        """
        List all users.
        :return: the list of users.
        """


def config_model() -> List[Json]:
    config_classes: Set[Type[Any]] = {ResotoUsersConfig}
    return dataclasses_to_resotocore_model(config_classes, use_optional_as_required=True)



import pytest

from resotocore.db import EstimatedSearchCost, EstimatedQueryCostRating
from resotocore.db.arango_query import to_query, query_cost, fulltext_term_combine
from resotocore.db.graphdb import GraphDB
from resotocore.db.model import QueryModel
from resotocore.model.model import Model
from resotocore.query.model import Query, Sort
from resotocore.query.query_parser import parse_query


def test_sort_order_for_synthetic_prop(foo_model: Model, graph_db: GraphDB) -> None:
    def check_sort_in_query(q: Query, expected_sort: str) -> None:
        query_str, _ = to_query(graph_db, QueryModel(q, foo_model))
        assert f"SORT {expected_sort}" in query_str, f"Expected {expected_sort} in {query_str}"

    check_sort_in_query(Query.by("foo").add_sort(Sort("reported.age")), "m0.reported.ctime desc")
    check_sort_in_query(Query.by("foo").add_sort(Sort("some.age")), "m0.some.age asc")
    check_sort_in_query(Query.by("foo").add_sort(Sort("reported.ctime")), "m0.reported.ctime asc")
    check_sort_in_query(Query.by("foo").add_sort(Sort("metadata.expired")), "m0.metadata.expired asc")


@pytest.mark.asyncio
async def test_query_cost(foo_model: Model, graph_db: GraphDB) -> None:
    async def cost(query_str: str) -> EstimatedSearchCost:
        query = parse_query(query_str)
        return await query_cost(graph_db, QueryModel(query, foo_model), False)

    c1 = await cost("aggregate(sum(1) as count):is(base) sort count asc")
    assert c1.full_collection_scan is False
    assert c1.rating is EstimatedQueryCostRating.simple

    c2 = await cost("is(base) sort count asc")
    assert c2.full_collection_scan is False
    assert c2.rating is EstimatedQueryCostRating.simple

    c3 = await cost("all sort reported.name asc")
    assert c3.full_collection_scan is True
    assert c3.rating is EstimatedQueryCostRating.complex

    c4 = await cost("all {parents: <-[0:]-} sort reported.name asc")
    assert c4.full_collection_scan is True
    assert c4.rating is EstimatedQueryCostRating.bad


def test_fulltext_term() -> None:
    part = parse_query('(a>0 and ("foo" and (b>1 and c>2 and "d")))').parts[0]
    ft, remaining = fulltext_term_combine(part.term)
    assert str(remaining) == "((b > 1 and c > 2) and a > 0)"
    assert str(ft) == '("d" and "foo")'
    # there are 2 fulltext terms or combined with something else
    ft, remaining = fulltext_term_combine(parse_query('(a>0 and "b") or ("c" and "d")').parts[0].term)
    assert ft is None  # fulltext index can not be utilized
    ft, remaining = fulltext_term_combine(parse_query('a>0 {c: <--} "fulltext"').parts[0].term)
    assert ft is None  # fulltext index can not be utilized
    ft, remaining = fulltext_term_combine(parse_query('a>0 {c: <-- "fulltext" }').parts[0].term)
    assert ft is None  # fulltext index can not be utilized
    ft, remaining = fulltext_term_combine(parse_query('"a" and "b" or "c" and "d"').parts[0].term)
    assert str(ft) == '((("a" and "b") or "c") and "d")'


def test_fulltext_index_query(foo_model: Model, graph_db: GraphDB) -> None:
    def query_string(query: str) -> str:
        query_str, _ = to_query(graph_db, QueryModel(parse_query(query), foo_model))
        return query_str

    single_ft_index = (
        "LET m0=(FOR ft in search_ns SEARCH ANALYZER(PHRASE(ft.flat, @b0), 'delimited') "
        "SORT BM25(ft) DESC RETURN ft) "
        'FOR result in m0 RETURN UNSET(result, ["flat"])'
    )
    assert query_string('"a"') == single_ft_index
    assert query_string('"some other fulltext string"') == single_ft_index
    # and/or is combined correctly
    assert (
        "ANALYZER((((PHRASE(ft.flat, @b0)) and (PHRASE(ft.flat, @b1))) or "
        "(PHRASE(ft.flat, @b2))) and (PHRASE(ft.flat, @b3)), 'delimited')"
    ) in query_string('"a" and "b" or "c" and "d"')


def test_ancestors_kind_lookup(foo_model: Model, graph_db: GraphDB) -> None:
    # 1234 is coerced to a string
    query = "ancestors.account.reported.name==1234"
    assert to_query(graph_db, QueryModel(parse_query(query), foo_model))[1] == {"b0": "1234"}


def test_escape_property_path(foo_model: Model, graph_db: GraphDB) -> None:
    raw = "metadata.replace.with.filter.sort.bla==true"
    query = to_query(graph_db, QueryModel(parse_query(raw), foo_model))[0]
    # aql keywords are escaped with backslashes
    assert "m0.metadata.`replace`.`with`.`filter`.`sort`.bla" in query


def test_with_query_with_limit(foo_model: Model, graph_db: GraphDB) -> None:
    query = "is(foo) with(empty, -->) limit 2"
    query_str, _ = to_query(graph_db, QueryModel(parse_query(query), foo_model))
    # make sure, there is no limit in the filter statement
    assert "LET filter0 = (FOR m0 in `ns` FILTER @b0 IN m0.kinds  RETURN m0)" in query_str
    # make sure the limit is applied to the with statement
    assert "FILTER counter1==1  LIMIT 0, 2 RETURN l0_l0_res" in query_str


def test_context(foo_model: Model, graph_db: GraphDB) -> None:
    query = 'is(foo) and nested[*].{name=true and inner[*].{name=true}} and parents[*].{some_int="23"}'
    aql, bind_vars = to_query(graph_db, QueryModel(parse_query(query).on_section("reported"), foo_model))
    # query unfolds all nested loops
    assert aql == (
        "LET filter0 = (LET nested_distinct0 = (FOR m0 in `ns`  FOR pre0 IN TO_ARRAY(m0.reported.nested) "
        "FOR pre1 IN TO_ARRAY(pre0.inner)  "
        "FOR pre2 IN TO_ARRAY(m0.reported.parents) "
        "FILTER ((@b0 IN m0.kinds) and ((pre0.name == @b1) and (pre1.name == @b2))) and (pre2.some_int == @b3) "
        "RETURN DISTINCT m0) FOR m1 in nested_distinct0  "
        'RETURN m1) FOR result in filter0 RETURN UNSET(result, ["flat"])'
    )
    # coercing works correctly for context terms
    assert bind_vars["b1"] == "true"  # true is coerced to a string
    assert bind_vars["b2"] == "true"  # inner true is coerced to a string
    assert bind_vars["b3"] == 23  # 23 is coerced to an int

    # fixed index works as well
    query = "is(foo) and inner[1].{name=true and inner[0].name==true}"
    aql, bind_vars = to_query(graph_db, QueryModel(parse_query(query).on_section("reported"), foo_model))
    assert aql == (
        "LET filter0 = (FOR m0 in `ns` FILTER (@b0 IN m0.kinds) and "
        "((m0.reported.inner[1].name == @b1) and (m0.reported.inner[1].inner[0].name == @b2))  RETURN m0) "
        'FOR result in filter0 RETURN UNSET(result, ["flat"])'
    )


def test_usage(foo_model: Model, graph_db: GraphDB) -> None:
    q, b = to_query(graph_db, QueryModel(parse_query("with_usage(3w, cpu, mem) is(foo)"), foo_model))
    assert q == (
        "LET filter0 = (FOR m0 in `ns` FILTER @b0 IN m0.kinds  RETURN m0)\n"
        "let with_usage0 = (\n"
        "    for r in filter0\n"
        "        let resource=r\n"
        "        let resource_usage = first(\n"
        "            for m in ns_usage\n"
        "            filter m.at>=@b1 and m.at<=@b2 and m.id==r._key\n"
        "            collect aggregate cpu_min = MIN(m.v.cpu.min), cpu_avg = AVG(m.v.cpu.avg), cpu_max = MAX(m.v.cpu.max), mem_min = MIN(m.v.mem.min), mem_avg = AVG(m.v.mem.avg), mem_max = MAX(m.v.mem.max), count = sum(1)\n"  # noqa: E501
        "            return {usage:{cpu: {min: cpu_min, avg: cpu_avg, max: cpu_max},mem: {min: mem_min, avg: mem_avg, max: mem_max},entries:count,start:@b3,duration:@b4}}\n"  # noqa: E501
        "        )\n"
        "        return resource_usage.usage.entries ? merge(resource, resource_usage) : resource\n"
        ")\n"
        ' FOR result in with_usage0 RETURN UNSET(result, ["flat"])'
    )



from dataclasses import dataclass
from enum import Enum
from abc import ABC, abstractmethod
import random
from typing import List, FrozenSet, Deque, Tuple
from datetime import datetime
from collections import deque


@dataclass(frozen=True)
class TipOfTheDay:
    description: str
    command_line: str
    difficulty: int


class SuggestionPolicy(Enum):
    RANDOM = 1
    NON_REPEATING = 2
    DAILY = 3


class SuggestionStrategy(ABC):
    def __init__(self, cloud_providers: FrozenSet[str]):
        self._cloud_providers = cloud_providers
        tip_builder = list(generic_tips)
        if "aws" in self._cloud_providers:
            tip_builder.extend(aws_tips)
        if "digitalocean" in self._cloud_providers:
            tip_builder.extend(digitalocean_tips)
        if "k8s" in self._cloud_providers:
            tip_builder.extend(k8s_tips)
        if "gcp" in self._cloud_providers:
            tip_builder.extend(gcp_tips)
        tip_builder = sorted(tip_builder, key=lambda x: x.difficulty)
        self._tips: Tuple[TipOfTheDay, ...] = tuple(tip_builder)

    @abstractmethod
    def suggest(self) -> TipOfTheDay:
        pass


class RandomSuggestionStrategy(SuggestionStrategy):
    """
    Makes a random suggestion from the list of tips.
    """

    def suggest(self) -> TipOfTheDay:
        return random.sample(self._tips, 1)[0]


class NonRepeatingSuggestionStrategy(SuggestionStrategy):
    """
    Makes a random suggestion from the list of tips, but does not repeat
    a suggestion until all tips have been suggested.

    Keeps track of the state based on the session id.
    """

    def __init__(self, cloud_providers: FrozenSet[str]):
        super().__init__(cloud_providers)
        self._next_tips: Deque[TipOfTheDay] = deque()

    def suggest(self) -> TipOfTheDay:
        if not self._next_tips:
            self._next_tips.extend(self._tips)
        tip = self._next_tips.popleft()
        return tip


class DailyStrategy(SuggestionStrategy):
    """
    Makes a suggestion based on the day of the year. The tip is determined
    by the day of the year modulo the number of tips.
    """

    def suggest(self) -> TipOfTheDay:
        return self._tips[datetime.now().timetuple().tm_yday % len(self._tips)]


def get_suggestion_strategy(policy: SuggestionPolicy, configured_clouds: FrozenSet[str]) -> SuggestionStrategy:
    if policy == SuggestionPolicy.RANDOM:
        return RandomSuggestionStrategy(configured_clouds)
    elif policy == SuggestionPolicy.NON_REPEATING:
        return NonRepeatingSuggestionStrategy(configured_clouds)
    elif policy == SuggestionPolicy.DAILY:
        return DailyStrategy(configured_clouds)
    else:
        raise NotImplementedError


generic_tips = [
    TipOfTheDay(
        description='Find anything that contains "10.0.0.199"',
        command_line='search "10.0.0.199"',
        difficulty=1,
    ),
    TipOfTheDay(
        description="Count your instances by kind",
        command_line="search is(instance) | count kind",
        difficulty=1,
    ),
    TipOfTheDay(
        description="Format the result of your search as a json and write it to a file",
        command_line="search is(instance) | format --json | write instances.json",
        difficulty=2,
    ),
    TipOfTheDay(
        description="Find first 10 volumes, get their id and size, and display them as a markdown table",
        command_line="search is(volume) limit 10 | list id, volume_size --markdown",
        difficulty=4,
    ),
]

aws_tips: List[TipOfTheDay] = [
    TipOfTheDay(
        description="AWS EBS Volumes older than 90 days that had no I/O in the past 30 days",
        command_line="search is(aws_ec2_volume) and age > 90d and last_access > 30d",
        difficulty=2,
    ),
    TipOfTheDay(
        description="AWS EC2 instances that are missing the 'owner' tag",
        command_line="search is(aws_ec2_instance) and not has_key(tags, owner)",
        difficulty=3,
    ),
    TipOfTheDay(
        description="AWS EC2 instances that are missing the 'owner' tag",
        command_line="search is(aws_ec2_instance) and tags.owner == null",
        difficulty=3,
    ),
    TipOfTheDay(
        description='AWS network interfaces that contain "10.0.0.199" in any field',
        command_line='search is(aws_ec2_network_interface) and "10.0.0.199"',
        difficulty=2,
    ),
    TipOfTheDay(
        description="Find AWS EC2 instances and merge the instance type information into the instance data",
        command_line="search is(aws_ec2_instance) {instance_type: <-- is(aws_ec2_instance_type)}",
        difficulty=5,
    ),
    TipOfTheDay(
        description="See the detailed information about a random AWS instance",
        command_line="search is(aws_ec2_instance) | tail 1 | dump",
        difficulty=1,
    ),
    TipOfTheDay(
        description="ALBs that have no target group",
        command_line="search is(aws_alb) with(empty, --> is(aws_alb_target_group)",
        difficulty=4,
    ),
    TipOfTheDay(
        description="ELBs without EC2 instances behind them",
        command_line="search is(aws_elb) with(empty, --> is(aws_ec2_instance))",
        difficulty=4,
    ),
    TipOfTheDay(
        description="Find all expired IAM server certificates",
        command_line="search is(aws_iam_server_certificate) and expires < '@now@'",
        difficulty=3,
    ),
    TipOfTheDay(
        description="Find AWS ALB target groups of instance type that have no instances",
        command_line="search is(aws_alb_target_group) and "
        "target_type == instance with(empty, --> is(aws_ec2_instance))",
        difficulty=5,
    ),
]

gcp_tips: List[TipOfTheDay] = [
    TipOfTheDay(
        description="See the detailed information about a random GCP instance",
        command_line="search is(gcp_instance) | tail 1 | dump",
        difficulty=2,
    ),
]

digitalocean_tips: List[TipOfTheDay] = [
    TipOfTheDay(
        description="Digitalocean Volumes older than 90 days that had no I/O in the past 30 days",
        command_line="search is(digitalocean_volume) and age > 90d and last_access > 30d",
        difficulty=2,
    ),
    TipOfTheDay(
        description="DigitalOcean droplets that are missing the 'owner' tag",
        command_line="search is(digitalocean_droplet) and not has_key(tags, owner)",
        difficulty=3,
    ),
    TipOfTheDay(
        description="DigitalOcean droplets that are missing the 'owner' tag",
        command_line="search is(digitalocean_droplet) and tags.owner == null",
        difficulty=3,
    ),
    TipOfTheDay(
        description='DigitalOcean VPCs that contain "10.0.0.199" in any field',
        command_line='search is(digitalocean_vpc) and "10.0.0.199"',
        difficulty=2,
    ),
    TipOfTheDay(
        description="See the detailed information about a random DigitalOcean droplet",
        command_line="search is(digitalocean_droplet) | tail 1 | dump",
        difficulty=2,
    ),
    TipOfTheDay(
        description="Find all expired DigitalOceans certificates",
        command_line="search is(digitalocean_certificate) and expires < '@now@'",
        difficulty=3,
    ),
]

k8s_tips: List[TipOfTheDay] = [
    TipOfTheDay(
        description="Kubernetes pods that are missing the 'owner' tag",
        command_line="search is(k8s_pod) and not has_key(labels, owner)",
        difficulty=3,
    ),
]



import json
import logging
import re
from collections import defaultdict
from typing import AsyncGenerator, List, Dict, AsyncIterator, Tuple, Callable, Optional, Any

import yaml
from aiohttp.web import StreamResponse, Request, Response, json_response
from networkx import DiGraph, cytoscape_data, generate_graphml

from resotocore.cli import is_node
from resotocore.constants import plain_text_blacklist
from resotocore.error import QueryTookToLongError
from resotocore.model.resolve_in_graph import NodePath
from resotocore.model.typed_model import to_json
from resotocore.types import Json, JsonElement
from resotocore.util import del_value_in_path, value_in_path, value_in_path_get, count_iterator, identity

log = logging.getLogger(__name__)


async def respond_json(gen: AsyncIterator[JsonElement], **json_args: Any) -> AsyncGenerator[str, None]:
    sep = ","
    yield "["
    first = True
    async for item in gen:
        js = json.dumps(to_json(item), **json_args)
        if not first:
            yield sep
        yield js
        first = False
    yield "]"


async def respond_ndjson(gen: AsyncIterator[JsonElement]) -> AsyncGenerator[str, None]:
    async for item in gen:
        js = json.dumps(to_json(item), check_circular=False)
        yield js


async def respond_yaml(gen: AsyncIterator[JsonElement]) -> AsyncGenerator[str, None]:
    flag = False
    sep = "---"
    async for item in gen:
        yml = yaml.dump(to_json(item), default_flow_style=False, sort_keys=False)
        if flag:
            yield sep
        yield yml
        flag = True


async def respond_dot(gen: AsyncIterator[JsonElement]) -> AsyncGenerator[str, None]:
    # We use the paired12 color scheme: https://graphviz.org/doc/info/colors.html with color names as 1-12
    cit = count_iterator()
    colors: Dict[str, int] = defaultdict(lambda: (next(cit) % 12) + 1)
    node = "node [shape=Mrecord colorscheme=paired12]"
    edge = "edge [arrowsize=0.5]"
    yield f"digraph {{\nrankdir=LR\noverlap=false\nsplines=true\n{node}\n{edge}"
    in_account: Dict[str, List[str]] = defaultdict(list)
    async for item in gen:
        if isinstance(item, dict):
            type_name = item.get("type")
            if type_name == "node":
                uid = value_in_path(item, NodePath.node_id)
                if uid:
                    name = re.sub("[^a-zA-Z\\-0-9]", "_", value_in_path_get(item, NodePath.reported_name, "n/a"))
                    kind = value_in_path_get(item, NodePath.reported_kind, "n/a")
                    account = value_in_path_get(item, NodePath.ancestor_account_name, "graph_root")
                    paired12 = colors[kind]
                    in_account[account].append(uid)
                    yield f' "{uid}" [label="{name}|{kind}", style=filled fillcolor={paired12}];'
            elif type_name == "edge":
                from_node = value_in_path(item, NodePath.from_node)
                to_node = value_in_path(item, NodePath.to_node)
                edge_type = value_in_path(item, NodePath.edge_type)
                if from_node and to_node:
                    yield f' "{from_node}" -> "{to_node}" [label="{edge_type}"]'
        else:
            raise AttributeError(f"Expect json object but got: {type(item)}: {item}")
    # All elements in the same account are rendered as dedicated subgraph
    for account, uids in in_account.items():
        yield f' subgraph "{account}" {{'
        for uid in uids:
            yield f'    "{uid}"'
        yield " }"

    yield "}"


async def respond_text(gen: AsyncIterator[JsonElement]) -> AsyncGenerator[str, None]:
    def filter_attrs(js: Json) -> Json:
        result: Json = js
        for path in plain_text_blacklist:
            del_value_in_path(js, path)
        return result

    def to_result(js: JsonElement) -> JsonElement:
        # if js is a node, the resulting content should be filtered
        return filter_attrs(js) if is_node(js) else js  # type: ignore

    try:
        flag = False
        sep = "---"
        async for item in gen:
            js = to_json(item)
            if isinstance(js, (dict, list)):
                if flag:
                    yield sep
                yml = yaml.dump(to_result(js), allow_unicode=True, default_flow_style=False, sort_keys=False)
                yield yml
            else:
                yield str(js)
            flag = True
    except QueryTookToLongError:
        yield (
            "\n\n---------------------------------------------------\n"
            "Query took too long.\n"
            "Try one of the following:\n"
            "- refine your query\n"
            "- add a limit to your query\n"
            "- define a longer timeout via env var search_timeout\n"
            "  e.g. $> search_timeout=60s query all\n"
            "---------------------------------------------------\n\n"
        )


async def result_to_graph(gen: AsyncIterator[JsonElement], render_node: Callable[[Json], Json] = identity) -> DiGraph:
    result = DiGraph()
    async for item in gen:
        if isinstance(item, dict):
            type_name = item.get("type")
            if type_name == "node":
                uid = value_in_path(item, NodePath.node_id)
                json_result = render_node(item)
                if uid:
                    result.add_node(uid, **json_result)
            elif type_name == "edge":
                from_node = value_in_path(item, NodePath.from_node)
                to_node = value_in_path(item, NodePath.to_node)
                if from_node and to_node:
                    result.add_edge(from_node, to_node)
        else:
            raise AttributeError(f"Expect json object but got: {type(item)}: {item}")
    return result


async def respond_cytoscape(gen: AsyncIterator[JsonElement]) -> AsyncGenerator[str, None]:
    # Note: this is a very inefficient way of creating a response, since it creates the graph in memory
    # on the server side, so we can reuse the networkx code.
    # This functionality can be reimplemented is a streaming way.
    graph = await result_to_graph(gen, lambda js: value_in_path_get(js, NodePath.reported, {}))
    yield json.dumps(cytoscape_data(graph))


async def respond_graphml(gen: AsyncIterator[JsonElement]) -> AsyncGenerator[str, None]:
    # Note: this is a very inefficient way of creating a response, since it creates the graph in memory
    # on the server side, so we can reuse the networkx code.
    # This functionality can be reimplemented is a streaming way.
    def no_nested_props(js: Json) -> Json:
        reported: Json = value_in_path_get(js, NodePath.reported, {})
        res = {k: v for k, v in reported.items() if v is not None and not isinstance(v, (dict, list))}
        return res

    graph = await result_to_graph(gen, no_nested_props)
    for line in generate_graphml(graph):
        yield line


async def result_string_gen(request: Request, gen: AsyncIterator[JsonElement]) -> Tuple[str, AsyncIterator[str]]:
    accept = request.headers.get("accept", "application/json")
    if accept in ["application/x-ndjson", "application/ndjson"]:
        return "application/x-ndjson", respond_ndjson(gen)
    elif accept == "application/json":
        return "application/json", respond_json(gen)
    elif accept in ["text/plain"]:
        return "text/plain", respond_text(gen)
    elif accept in ["application/yaml", "text/yaml"]:
        return "text/yaml", respond_yaml(gen)
    elif accept == "application/vnd.cytoscape+json":
        return "application/vnd.cytoscape+json", respond_cytoscape(gen)
    elif accept in ["application/graphml+xml", "application/vnd.graphml+xml"]:
        return "application/graphml+xml", respond_graphml(gen)
    elif accept.startswith("text/vnd.graphviz"):
        return "text/yaml", respond_dot(gen)
    else:
        return "application/json", respond_json(gen)


async def result_binary_gen(request: Request, gen: AsyncIterator[JsonElement]) -> Tuple[str, AsyncIterator[bytes]]:
    content_type, str_gen = await result_string_gen(request, gen)

    async def encode_utf8() -> AsyncIterator[bytes]:
        async for elem in str_gen:
            yield elem.encode("utf-8")

    return content_type, encode_utf8()


async def single_result(
    request: Request, js: JsonElement, headers: Optional[Dict[str, Optional[str]]] = None
) -> StreamResponse:
    non_empty = {k: v for k, v in headers.items() if isinstance(v, str)} if headers else None
    accept = request.headers.get("accept", "application/json")
    if accept in ["application/yaml", "text/yaml"]:
        yml = yaml.dump(js)
        return Response(text=yml, content_type="application/yaml", headers=non_empty)
    else:
        return json_response(js, headers=non_empty)



from pathlib import Path
from tempfile import TemporaryDirectory
from resotocore.config.config_override_service import ConfigOverrideService
import os
from resotocore.ids import ConfigId
import pytest
from typing import Optional, Any
from resotocore.types import Json
import asyncio
from resotocore.model.model import Model, ComplexKind, Property


@pytest.mark.asyncio
async def test_get_overrides() -> None:
    # create a temp file with a custom config
    with TemporaryDirectory() as tmp:
        # config with env var override
        resotocore_1_conf = Path(tmp, "resoto.core.yml")
        resotocore_1_conf.write_text(
            """
resotocore:
    api:
        web_hosts: ["11.12.13.14"]
        web_port: $(WEB_PORT)
        """,
            encoding="utf-8",
        )

        resotoworker_conf = Path(tmp, "resoto.worker.yml")
        resotoworker_conf.write_text(
            """
resotoworker:
    collector: ['digitalocean', '$(OTHER_COLLECTOR)']
            """
        )

        os.environ["WEB_PORT"] = "1337"

        async def get_configs_model() -> Model:
            return Model.empty()

        override_service = ConfigOverrideService([Path(tmp)], get_configs_model)
        await override_service.load()

        assert override_service.get_override(ConfigId("resoto.core")) == {
            "resotocore": {
                "api": {
                    "web_hosts": ["11.12.13.14"],
                    "web_port": "$(WEB_PORT)",
                }
            }
        }

        assert override_service.get_override(ConfigId("resoto.worker")) == {
            "resotoworker": {"collector": ["digitalocean", "$(OTHER_COLLECTOR)"]}
        }

        assert override_service.get_all_overrides() == {
            "resoto.core": {
                "resotocore": {
                    "api": {
                        "web_hosts": ["11.12.13.14"],
                        "web_port": "$(WEB_PORT)",
                    }
                }
            },
            "resoto.worker": {"resotoworker": {"collector": ["digitalocean", "$(OTHER_COLLECTOR)"]}},
        }


@pytest.mark.asyncio
async def test_hooks() -> None:
    # create a temp file with a custom config
    with TemporaryDirectory() as tmp:
        # config with env var override
        foo_conf = Path(tmp, "foo.yml")
        foo_conf.write_text(
            """
bar:
    foo: 42
        """,
            encoding="utf-8",
        )

        async def get_configs_model() -> Model:
            return Model.empty()

        override_service = ConfigOverrideService([Path(tmp)], get_configs_model, sleep_time=0.05)
        await override_service.load()
        await override_service.start()

        override_update: Optional[Json] = None

        async def update_hook(json: Any) -> None:
            nonlocal override_update
            override_update = json

        override_service.add_override_change_hook(update_hook)

        assert not override_update

        foo_conf.write_text(
            """
bar:
    foo: 1337
        """,
            encoding="utf-8",
        )

        await asyncio.sleep(0.25)  # wait for the watcher to update the config
        assert override_update == {"foo": {"bar": {"foo": 1337}}}


@pytest.mark.asyncio
async def test_validation() -> None:
    # create a temp file with a custom config
    with TemporaryDirectory() as tmp:
        # config with env var override
        resotocore_1_conf = Path(tmp, "resoto.core.yml")
        resotocore_1_conf.write_text(
            """
foo:
    bar: 42
    kind: foo""",
            encoding="utf-8",
        )

        foo = ComplexKind("foo", [], [Property("bar", "int32"), Property("kind", "string")])

        async def get_configs_model() -> Model:
            return Model.from_kinds([foo])

        override_service = ConfigOverrideService([Path(tmp)], get_configs_model)
        await override_service.load()

        assert override_service.get_override(ConfigId("resoto.core")) == {
            "foo": {
                "bar": 42,
                "kind": "foo",
            }
        }

        resotocore_1_conf.write_text(
            """
foo:
    bar: bar
    kind: foo""",
            encoding="utf-8",
        )

        override_service = ConfigOverrideService([Path(tmp)], get_configs_model)
        await override_service.load()

        # invalid config should be ignored
        assert override_service.get_override(ConfigId("resoto.core")) is None


@pytest.mark.asyncio
async def test_load_json() -> None:
    # create a temp file with a custom config
    with TemporaryDirectory() as tmp:
        # config with env var override
        resotocore_1_conf = Path(tmp, "resoto.core.json")
        resotocore_1_conf.write_text(
            """
{
    "resotocore": {
        "api": {
            "web_hosts": ["11.12.13.14"],
            "web_port": "$(WEB_PORT)"
        }
    }
}""",
            encoding="utf-8",
        )

        resotoworker_conf = Path(tmp, "resoto.worker.json")
        resotoworker_conf.write_text(
            """
{
    "resotoworker": {
        "collector": ["digitalocean", "$(OTHER_COLLECTOR)"]
    }
}"""
        )

        os.environ["WEB_PORT"] = "1337"

        async def get_configs_model() -> Model:
            return Model.empty()

        override_service = ConfigOverrideService([Path(tmp)], get_configs_model)
        await override_service.load()

        assert override_service.get_override(ConfigId("resoto.core")) == {
            "resotocore": {
                "api": {
                    "web_hosts": ["11.12.13.14"],
                    "web_port": "$(WEB_PORT)",
                }
            }
        }

        assert override_service.get_override(ConfigId("resoto.worker")) == {
            "resotoworker": {"collector": ["digitalocean", "$(OTHER_COLLECTOR)"]}
        }

        assert override_service.get_all_overrides() == {
            "resoto.core": {
                "resotocore": {
                    "api": {
                        "web_hosts": ["11.12.13.14"],
                        "web_port": "$(WEB_PORT)",
                    }
                }
            },
            "resoto.worker": {"resotoworker": {"collector": ["digitalocean", "$(OTHER_COLLECTOR)"]}},
        }



import asyncio
import logging
from asyncio import Task
from datetime import timedelta
from typing import Optional, List

from resotocore.message_bus import MessageBus, CoreMessage, Message
from resotocore.ids import SubscriberId
from resotocore.task import TaskHandler
from resotocore.task.task_description import PerformAction, Workflow
from resotocore.util import uuid_str

log = logging.getLogger(__name__)


def wait_and_start(
    workflows: List[Workflow],
    task_handler: TaskHandler,
    message_bus: MessageBus,
    wait_after_connect: timedelta = timedelta(seconds=10),
) -> Task[None]:
    """
    This function is used to trigger workflows automatically, when the related subscribing actor connects.
    Such behaviour can be useful during startup, when we do not want to wait until the next scheduled time triggers.
    Note: this is a one_off action: after the initial connect has happened, the task will complete.

    :param workflows: the known workflows.
    :param task_handler: the job handler to start the workflow.
    :param message_bus: the message bus to listen for incoming subscribers.
    :param wait_after_connect: amount of time to wait between connect and workflow trigger.
    :return: running task that will succeed once the first subscribing actor connects.
    """

    log.info("Wait for subscribing actors to start the related workflow")
    # get all action message types the workflows are waiting for
    wait_for = {s.action.message_type: wf for wf in workflows for s in wf.steps if isinstance(s.action, PerformAction)}

    def workflow_if_actor(msg: Message) -> Optional[Workflow]:
        channels: Optional[List[str]] = msg.data.get("channels")
        if channels:
            for ch in channels:
                if ch in wait_for:
                    return wait_for[ch]
        return None

    async def wait_for_subscriber() -> None:
        subscriber_id = SubscriberId(f"resotocore.wait_for_actor_{uuid_str()}")
        async with message_bus.subscribe(subscriber_id, [CoreMessage.Connected]) as bus:
            while True:
                message = await bus.get()
                maybe_workflow = workflow_if_actor(message)
                if maybe_workflow:
                    log.info(
                        f"Subscribed actor for workflow {maybe_workflow.name} connected. "
                        f"Wait for {wait_after_connect} seconds."
                    )
                    # wait before we start the workflow
                    await asyncio.sleep(wait_after_connect.total_seconds())
                    log.info(f"Start workflow {maybe_workflow.name}")
                    # start the workflow
                    await task_handler.start_task_by_descriptor_id(maybe_workflow.id)
                    # exit the loop and destroy the listener
                    break
        log.info("task done")

    return asyncio.create_task(wait_for_subscriber())



import pytest

from resotocore.db.system_data_db import SystemDataDb


@pytest.mark.asyncio
async def test_system_info(system_data_db: SystemDataDb) -> None:
    assert await system_data_db.info() is not None
    key, certificate = await system_data_db.ca()
    assert key is not None
    assert certificate is not None


@pytest.mark.asyncio
async def test_system_data_update(system_data_db: SystemDataDb) -> None:
    existing = await system_data_db.info()
    assert await system_data_db.update_info(company="foo") == {**existing, "company": "foo"}
    assert await system_data_db.update_info(test="bla") == {**existing, "company": "foo", "test": "bla"}
    assert await system_data_db.info() == {**existing, "company": "foo", "test": "bla"}



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, AsyncIterator, List, Callable, Awaitable

from attrs import define
from jsons import set_deserializer, set_serializer

from resotocore.ids import ConfigId
from resotocore.model.model import Model, Kind
from resotocore.types import Json
from resotocore.util import value_in_path, value_in_path_get


@define(order=True, hash=True, frozen=True)
class ConfigEntity:
    id: ConfigId
    config: Json
    revision: Optional[str] = None

    def analytics(self) -> Dict[str, Any]:
        analytics: Dict[str, Any] = {"config_id": self.id}
        if "resotoworker" not in self.config:
            return analytics

        # provider information
        collectors: List[str] = []
        collectors.extend(value_in_path_get(self.config, ["resotoworker", "collector"], []))
        if "example" in collectors:
            collectors.remove("example")
        analytics["how_many_providers"] = len(collectors)

        # authentication information
        if "aws" in collectors:
            aws_use_access_secret_key = (
                value_in_path(self.config, ["resotoworker", "aws", "access_key_id"]) is not None
                and value_in_path(self.config, ["resotoworker", "aws", "secret_access_key"]) is not None
            )
            aws_use_role = value_in_path(self.config, ["resotoworker", "aws", "role"]) is not None
            aws_use_profiles = value_in_path(self.config, ["resotoworker", "aws", "profiles"]) is not None
            aws_use_accounts = value_in_path(self.config, ["resotoworker", "aws", "account"]) is not None
            aws_use_scrape_org = value_in_path(self.config, ["resotoworker", "aws", "scrape_org"])
            analytics["aws"] = True
            analytics["aws_use_access_secret_key"] = aws_use_access_secret_key
            analytics["aws_use_role"] = aws_use_role
            analytics["aws_use_profiles"] = aws_use_profiles
            analytics["aws_use_accounts"] = aws_use_accounts
            analytics["aws_use_scrape_org"] = aws_use_scrape_org

        if "digitalocean" in collectors:
            do_has_tokens = bool(value_in_path_get(self.config, ["resotoworker", "digitalocean", "api_tokens"], []))
            analytics["digitalocean"] = True
            analytics["do_use_config"] = do_has_tokens

        if "gcp" in collectors:
            gcp_service_accounts: List[str] = value_in_path_get(
                self.config, ["resotoworker", "gcp", "service_account"], []
            )
            gcp_auto_discovery = any(s == "" for s in gcp_service_accounts)
            gcp_use_file = any(s != "" for s in gcp_service_accounts)
            analytics["gcp"] = True
            analytics["gcp_use_auto_discovery"] = gcp_auto_discovery
            analytics["gcp_use_file"] = gcp_use_file

        if "k8s" in collectors:
            k8s_has_cfg_files = bool(value_in_path_get(self.config, ["resotoworker", "k8s", "config_files"], []))
            k8s_has_cfgs = bool(value_in_path_get(self.config, ["resotoworker", "k8s", "configs"], []))
            analytics["k8s"] = True
            analytics["k8s_use_kubeconfig"] = k8s_has_cfg_files
            analytics["k8s_use_manual"] = k8s_has_cfgs

        return analytics

    # noinspection PyUnusedLocal
    @staticmethod
    def from_json(js: Json, _: type = object, **kwargs: object) -> ConfigEntity:
        if "id" in js and "config" in js:
            return ConfigEntity(js["id"], js["config"], js.get("_rev"))
        else:
            raise AttributeError(f"Can not parse a ConfigEntity from this json: {js}")

    # noinspection PyUnusedLocal
    @staticmethod
    def to_json(o: ConfigEntity, **kw_args: object) -> Json:
        return dict(id=o.id, config=o.config, _rev=o.revision)


@define(order=True, hash=True, frozen=True)
class ConfigValidation:
    id: str
    external_validation: bool = False


class ConfigHandler(ABC):
    @abstractmethod
    def list_config_ids(self) -> AsyncIterator[ConfigId]:
        pass

    @abstractmethod
    async def get_config(
        self, cfg_id: ConfigId, apply_overrides: bool = True, resolve_env_vars: bool = True
    ) -> Optional[ConfigEntity]:
        pass

    @abstractmethod
    async def put_config(self, cfg: ConfigEntity, *, validate: bool = True, dry_run: bool = False) -> ConfigEntity:
        pass

    @abstractmethod
    async def patch_config(self, cfg: ConfigEntity, *, validate: bool = True, dry_run: bool = False) -> ConfigEntity:
        pass

    @abstractmethod
    async def copy_config(self, from_cfg_id: ConfigId, to_cfg_id: ConfigId) -> Optional[ConfigEntity]:
        pass

    @abstractmethod
    async def delete_config(self, cfg_id: ConfigId) -> None:
        pass

    @abstractmethod
    async def get_configs_model(self) -> Model:
        pass

    @abstractmethod
    async def update_configs_model(self, kinds: List[Kind]) -> Model:
        pass

    @abstractmethod
    def list_config_validation_ids(self) -> AsyncIterator[str]:
        pass

    @abstractmethod
    async def get_config_validation(self, cfg_id: str) -> Optional[ConfigValidation]:
        pass

    @abstractmethod
    async def put_config_validation(self, validation: ConfigValidation) -> ConfigValidation:
        pass

    @abstractmethod
    async def config_yaml(self, cfg_id: ConfigId, revision: bool = False) -> Optional[str]:
        pass


class ConfigOverride(ABC):
    @abstractmethod
    def add_override_change_hook(self, hook: Callable[[Dict[ConfigId, Json]], Awaitable[Any]]) -> None:
        pass

    @abstractmethod
    def get_override(self, config_id: ConfigId) -> Optional[Json]:
        pass

    @abstractmethod
    def get_all_overrides(self) -> Dict[ConfigId, Json]:
        pass

    @abstractmethod
    async def start(self) -> None:
        pass

    @abstractmethod
    async def stop(self) -> None:
        pass


# register serializer for this class
set_deserializer(ConfigEntity.from_json, ConfigEntity)
set_serializer(ConfigEntity.to_json, ConfigEntity)



import asyncio
from datetime import timedelta
from typing import List

import pytest
from arango.database import StandardDatabase

from resotocore.analytics import InMemoryEventSender
from resotocore.db import jobdb
from resotocore.db.async_arangodb import AsyncArangoDB
from resotocore.db.entitydb import EventEntityDb
from resotocore.db.jobdb import JobDb, EventJobDb
from resotocore.ids import TaskDescriptorId
from resotocore.task.task_description import Job, ExecuteCommand, EventTrigger


@pytest.fixture
async def job_sdb(test_db: StandardDatabase) -> JobDb:
    async_db = AsyncArangoDB(test_db)
    job_db = jobdb.job_db(async_db, "jobs")
    await job_db.create_update_schema()
    await job_db.wipe()
    return job_db


@pytest.fixture
def event_db(job_sdb: JobDb, event_sender: InMemoryEventSender) -> EventJobDb:
    return EventEntityDb(job_sdb, event_sender, "job")


@pytest.fixture
def jobs() -> List[Job]:
    wait = (EventTrigger("wait"), timedelta(seconds=30))
    return [
        Job(TaskDescriptorId("id1"), ExecuteCommand("echo hello"), timedelta(seconds=10), EventTrigger("run_job")),
        Job(TaskDescriptorId("id2"), ExecuteCommand("sleep 10"), timedelta(seconds=10), EventTrigger("run_job"), wait),
    ]


def job_id(job: Job) -> str:
    return job.id


@pytest.mark.asyncio
async def test_load(job_sdb: JobDb, jobs: List[Job]) -> None:
    await job_sdb.update_many(jobs)
    loaded = [sub async for sub in job_sdb.all()]
    assert jobs.sort(key=job_id) == loaded.sort(key=job_id)


@pytest.mark.asyncio
async def test_update(job_sdb: JobDb, jobs: List[Job]) -> None:
    # multiple updates should work as expected
    await job_sdb.update_many(jobs)
    await job_sdb.update_many(jobs)
    await job_sdb.update_many(jobs)
    loaded = [sub async for sub in job_sdb.all()]
    assert jobs.sort(key=job_id) == loaded.sort(key=job_id)


@pytest.mark.asyncio
async def test_delete(job_sdb: JobDb, jobs: List[Job]) -> None:
    await job_sdb.update_many(jobs)
    remaining = list(jobs)
    for _ in jobs:
        sub = remaining.pop()
        await job_sdb.delete_value(sub)
        loaded = [sub async for sub in job_sdb.all()]
        assert remaining.sort(key=job_id) == loaded.sort(key=job_id)
    assert len([sub async for sub in job_sdb.all()]) == 0


@pytest.mark.asyncio
async def test_events(event_db: EventJobDb, jobs: List[Job], event_sender: InMemoryEventSender) -> None:
    # 2 times update
    await event_db.update_many(jobs)
    await event_db.update_many(jobs)
    # 2 times delete
    for sub in jobs:
        await event_db.delete_value(sub)
    # make sure all events will arrive
    await asyncio.sleep(0.1)
    # ensure the correct count and order of events
    assert [a.kind for a in event_sender.events] == ["job-updated-many"] * 2 + ["job-deleted"] * 2



import argparse
import logging
import multiprocessing as mp
import os.path
import sys
from argparse import Namespace
from collections import namedtuple
from pathlib import Path
from typing import Optional, List, Callable, Tuple

import psutil
from arango.database import StandardDatabase
from parsy import Parser

from resotocore import async_extensions, version
from resotocore.analytics import AnalyticsEventSender
from resotocore.core_config import CoreConfig, parse_config, git_hash_from_file, inside_docker
from resotocore.db.db_access import DbAccess
from resotocore.model.adjust_node import DirectAdjuster
from resotocore.types import JsonElement
from resotocore.util import utc
from resotolib.args import ArgumentParser
from resotolib.jwt import add_args as jwt_add_args
from resotolib.logger import setup_logger
from resotolib.parse_util import make_parser, variable_p, equals_p, comma_p, json_value_dp
from resotolib.utils import iec_size_format, get_local_tzinfo

log = logging.getLogger(__name__)

SystemInfo = namedtuple(
    "SystemInfo",
    ["version", "git_hash", "cpus", "mem_available", "mem_total", "inside_docker", "time_zone", "started_at"],
)
started_at = utc()


@make_parser
def path_json_value_parser() -> Parser:
    key = yield variable_p
    yield equals_p
    value = yield json_value_dp.sep_by(comma_p, min=1)
    return key, value


def system_info() -> SystemInfo:
    mem = psutil.virtual_memory()
    return SystemInfo(
        version=version(),
        git_hash=git_hash_from_file() or "n/a",
        cpus=mp.cpu_count(),
        mem_available=iec_size_format(mem.available),
        mem_total=iec_size_format(mem.total),
        inside_docker=inside_docker(),
        time_zone=get_local_tzinfo().key,
        started_at=started_at,
    )


def parse_args(args: Optional[List[str]] = None) -> Namespace:
    def is_file(message: str) -> Callable[[str], str]:
        def check_file(path: str) -> str:
            if os.path.isfile(path):
                return path
            else:
                raise AttributeError(f"{message}: path {path} is not a directory!")

        return check_file

    def key_value(kv: str) -> Tuple[str, JsonElement]:
        try:
            key, value = path_json_value_parser.parse(kv)
            return (key, value[0]) if len(value) == 1 else (key, value)
        except Exception as ex:
            raise AttributeError(f"Can not parse config option: {kv}. Reason: {ex}") from ex

    parser = ArgumentParser(
        env_args_prefix="RESOTOCORE_",
        description="Maintains graphs of resources of any shape.",
        epilog="Keeps all the things.",
    )
    jwt_add_args(parser)
    parser.add_argument(
        "--graphdb-server",
        default="http://localhost:8529",
        dest="graphdb_server",
        help="Graph database server (default: http://localhost:8529)",
    )
    parser.add_argument(
        "--graphdb-database", default="resoto", dest="graphdb_database", help="Graph database name (default: resoto)"
    )
    parser.add_argument(
        "--graphdb-username", default="resoto", dest="graphdb_username", help="Graph database login (default: resoto)"
    )
    parser.add_argument(
        "--graphdb-password", default="", dest="graphdb_password", help='Graph database password (default: "")'
    )
    parser.add_argument(
        "--graphdb-root-password",
        default="",
        dest="graphdb_root_password",
        help="Graph root database password used for creating user and database if not existent.",
    )
    parser.add_argument(
        "--graphdb-bootstrap-do-not-secure",
        default=False,
        action="store_true",
        dest="graphdb_bootstrap_do_not_secure",
        help="Leave an empty root password during system setup process.",
    )
    parser.add_argument(
        "--graphdb-type", default="arangodb", dest="graphdb_type", help="Graph database type (default: arangodb)"
    )
    parser.add_argument(
        "--graphdb-no-ssl-verify",
        action="store_true",
        dest="graphdb_no_ssl_verify",
        help="If the connection should not be verified (default: False)",
    )
    parser.add_argument(
        "--graphdb-request-timeout",
        type=int,
        default=900,
        dest="graphdb_request_timeout",
        help="Request timeout in seconds (default: 900)",
    )
    parser.add_argument("--no-tls", default=False, action="store_true", help=argparse.SUPPRESS)
    parser.add_argument(
        "--cert",
        type=is_file("can not parse --cert"),
        dest="cert",
        help="Path to a single file in PEM format containing the host certificate. "
        "If no certificate is provided, it is created using the CA.",
    )
    parser.add_argument(
        "--cert-key",
        type=is_file("can not parse --cert-key"),
        dest="cert_key",
        help="In case a --cert is provided. Path to a file containing the private key.",
    )
    parser.add_argument(
        "--cert-key-pass",
        dest="cert_key_pass",
        type=str,
        help="In case a --cert is provided. Optional password to decrypt the private key file.",
    )
    parser.add_argument(
        "--ca-cert",
        type=is_file("can not parse --ca-cert"),
        dest="ca_cert",
        help="Path to a single file in PEM format containing the CA certificate.",
    )
    parser.add_argument(
        "--ca-cert-key",
        type=is_file("can not parse --ca-cert-key"),
        dest="ca_cert_key",
        help="Path to a file containing the private key for the CA certificate. "
        "New certificates can be created when a CA certificate and private key is provided. "
        "Without the private key, the CA certificate is only used for outgoing http requests.",
    )
    parser.add_argument(
        "--ca-cert-key-pass",
        dest="ca_cert_key_pass",
        type=str,
        help="Optional password to decrypt the private ca-cert-key file.",
    )
    parser.add_argument("--version", action="store_true", help="Print the version of resotocore and exit.")
    parser.add_argument(
        "--override",
        "-o",
        nargs="+",
        type=key_value,
        dest="config_override",
        default=[],
        help="Override configuration parameters. Format: path.to.property=value. "
        "The existing configuration will be patched with the provided values. "
        "A value can be a simple value or a comma separated list of values if a list is required. "
        "Note: this argument allows multiple overrides separated by space. "
        "Example: --override resotocore.api.web_hosts=localhost,some.domain resotocore.api.https_port=12345",
    )
    parser.add_argument(
        "--override-path",
        nargs="+",
        type=Path,
        dest="config_override_path",
        default=[],
        help="Override configuration parameters via a YAML file or directory with YAML files. "
        "The existing configuration will be patched with the provided values. "
        "Note: this argument allows multiple overrides separated by space, in this case the "
        "resulting configuration will be the merge of all the provided files, in the order they are provided."
        "The same section can be overridden multiple times, in this case the last override will be used."
        "Example: --override-path /path/to/config/dir/ /path/to/your/config.yaml"
        "Be sure to specify the correct config id in the yaml, e.g. override for resotoworker would looke like:\n"
        """
resoto.worker:
    resotoworker:
        ...
    aws:
        ...""",
    )
    parser.add_argument(
        "--verbose", "-v", dest="verbose", default=False, action="store_true", help="Enable verbose logging."
    )
    parser.add_argument(  # No default here on purpose: it can be reconfigured!
        "--debug", default=None, action="store_true", help="Enable debug mode. If not defined use configuration."
    )
    parser.add_argument("--ui-path", help=argparse.SUPPRESS)  # no effect any longer, only here to not break anything
    parser.add_argument("--analytics-opt-out", default=None, action="store_true", help=argparse.SUPPRESS)

    parsed: Namespace = parser.parse_args(args if args else [])

    if parsed.version:
        # print here on purpose, since logging is not set up yet.
        print(f"resotocore {version()}")
        sys.exit(0)

    return parsed


def empty_config(args: Optional[List[str]] = None) -> CoreConfig:
    return parse_config(parse_args(args or []), {}, lambda: None)


# Note: this method should be called from every started process as early as possible
def setup_process(args: Namespace, config: Optional[CoreConfig] = None) -> None:
    if config:
        configure_logging(config.runtime.log_level, (config.runtime.debug or False) | (args.verbose or False))
    else:
        configure_logging("info", (args.debug or False) | (args.verbose or False))
    # set/reset process creation method
    reset_process_start_method()
    # reset global async thread pool (forked processes need to create a fresh pool)
    async_extensions.GlobalAsyncPool = None


def reconfigure_logging(config: CoreConfig) -> None:
    configure_logging(config.runtime.log_level, (config.runtime.debug or False) | (config.args.verbose or False))


def configure_logging(log_level: str, verbose: bool) -> None:
    # Note: if another appender than the log appender is used, proper multiprocess logging needs to be enabled.
    # See https://docs.python.org/3/howto/logging-cookbook.html#logging-to-a-single-file-from-multiple-processes
    level = log_level.upper()
    setup_logger("resotocore", level=level, verbose=verbose, force=True)

    # adjust log levels for specific loggers
    if verbose:
        logging.getLogger("resoto").setLevel(logging.DEBUG)
        logging.getLogger("resotolib").setLevel(logging.DEBUG)
        logging.getLogger("resotocore").setLevel(logging.DEBUG)
        # in case of restart: reset the original level
        logging.getLogger("posthog").setLevel(level)
        logging.getLogger("backoff").setLevel(level)
        logging.getLogger("transitions.core").setLevel(level)
        logging.getLogger("apscheduler.executors").setLevel(level)
        logging.getLogger("apscheduler.scheduler").setLevel(level)
    else:
        # in case of restart: reset the original level
        logging.getLogger("resoto").setLevel(level)
        logging.getLogger("resotolib").setLevel(level)
        logging.getLogger("resotocore").setLevel(level)
        # mute analytics transmission errors unless debug is enabled
        logging.getLogger("posthog").setLevel(logging.FATAL)
        logging.getLogger("backoff").setLevel(logging.FATAL)
        # transitions (fsm) creates a lot of log noise. Only show warnings.
        logging.getLogger("transitions.core").setLevel(logging.WARNING)
        # apscheduler uses the term Job when it triggers, which confuses people.
        logging.getLogger("apscheduler.executors").setLevel(logging.WARNING)
        logging.getLogger("apscheduler.scheduler").setLevel(logging.WARNING)


def reset_process_start_method() -> None:
    preferred = "spawn"
    current = mp.get_start_method(True)
    if current != preferred:
        if preferred in mp.get_all_start_methods():
            log.debug(f"Set process start method to {preferred}")
            mp.set_start_method(preferred, True)
            return
        log.warning(f"{preferred} method not available. Have {mp.get_all_start_methods()}. Use {current}")


def db_access(config: CoreConfig, db: StandardDatabase, event_sender: AnalyticsEventSender) -> DbAccess:
    adjuster = DirectAdjuster()
    return DbAccess(db, event_sender, adjuster, config)
